def replace_subarray(arr, start, num_elms, value=0, replace_arr=None):
    """Replaces a subsection within an array with another
    set of values.

    :param arr: The array to replace values within
    :type arr: list
    :param start: The starting index for replacement
    :type start: int
    :param num_elms: The number of elements to be replaced
    :type num_elms: int
    :param value: The value to replace elements within the
    subarray with, defaults to 0
    :type value: any, optional
    :param replace_arr: A subarray to insert within
    the passed array, defaults to None
    :type replace_arr: list, optional
    """

    if replace_arr:
        arr[start:start + num_elms] = replace_arr
    else:
        arr[start:start + num_elms] = [value] * num_elms


def format_message(data, split, name):
    """Formats a given byte message in hex format split
    into payload and subcommand sections.

    :param data: A series of bytes
    :type data: bytes
    :param split: The location of the payload/subcommand split
    :type split: integer
    :param name: The name featured in the start/end messages
    :type name: string
    :return: The formatted data
    :rtype: string
    """

    payload = ""
    subcommand = ""
    for i in range(0, len(data)):
        data_byte = str(hex(data[i]))[2:].upper()
        if len(data_byte) < 2:
            data_byte = "0" + data_byte
        if i <= split:
            payload += "0x" + data_byte + " "
        else:
            subcommand += "0x" + data_byte + " "

    formatted = (
        f"--- {name} Msg ---\n" +
        f"Payload:    {payload}\n" +
        f"Subcommand: {subcommand}")

    return formatted


def format_msg_controller(data):
    """Prints a formatted message from a controller

    :param data: The bytes from the controller message
    :type data: bytes
    """

    return format_message(data, 13, "Controller")


def format_msg_switch(data):
    """Prints a formatted message from a Switch

    :param data: The bytes from the Switch message
    :type data: bytes
    """

    return format_message(data, 10, "Switch")



import enum


class Controller(enum.Enum):
    JOYCON_L = 0x01
    JOYCON_R = 0x02
    PRO_CONTROLLER = 0x03

    def device_name(self):
        """
        :returns corresponding bluetooth device name
        """
        if self == Controller.JOYCON_L:
            return 'Joy-Con (L)'
        elif self == Controller.JOYCON_R:
            return 'Joy-Con (R)'
        elif self == Controller.PRO_CONTROLLER:
            return 'Pro Controller'
        else:
            raise NotImplementedError()

    @staticmethod
    def from_arg(arg):
        if arg == 'JOYCON_R':
            return Controller.JOYCON_R
        elif arg == 'JOYCON_L':
            return Controller.JOYCON_L
        elif arg == 'PRO_CONTROLLER':
            return Controller.PRO_CONTROLLER
        else:
            raise ValueError(f'Unknown controller "{arg}".')



from .server import ControllerServer
from .controller import ControllerTypes
from .controller import Controller
from .protocol import ControllerProtocol
from .protocol import SwitchReportParser
from .protocol import SwitchResponses




delay_override = False
delay = 1/15

async def debug(*args):
    global delay_override
    global delay
    if len(args) > 0:
        delay_override = True
        delay = 1/float(args[0])

def get_delay(old):
    return delay if delay_override else old



import logging
from datetime import datetime


def create_logger(debug=False, log_to_file=False, disable_logging=False):

    logger = logging.getLogger('nxbt')

    if disable_logging:
        null_handler = logging.NullHandler()
        logger.addHandler(null_handler)
        return logger

    if debug:
        logger.setLevel(logging.DEBUG)

    if log_to_file:
        file_handler = logging.FileHandler(f'./nxbt {datetime.now()}.log')
        file_handler.setFormatter(
            logging.Formatter("[%(asctime)s] %(levelname)s in %(module)s: %(message)s")
        )
        logger.addHandler(file_handler)
    else:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(
            logging.Formatter("[%(asctime)s] %(levelname)s in %(module)s: %(message)s")
        )
        logger.addHandler(stream_handler)

    return logger



from enum import Enum

from joycontrol.controller import Controller


class InputReport:
    """
    Class to create Input Reports. Reference:
    https://github.com/dekuNukem/Nintendo_Switch_Reverse_Engineering/blob/master/bluetooth_hid_notes.md
    """
    def __init__(self, data=None):
        if not data:
            self.data = [0x00] * 364
            # all input reports are prepended with 0xA1
            self.data[0] = 0xA1
        else:
            if data[0] != 0xA1:
                raise ValueError('Input reports must start with 0xA1')
            self.data = data

    def clear_sub_command(self):
        """
        Clear sub command reply data of 0x21 input reports
        """
        for i in range(14, 51):
            self.data[i] = 0x00

    def get_stick_data(self):
        # TODO: Not every input report has stick data
        return self.data[7:13]

    def get_sub_command_reply_data(self):
        if len(self.data) < 50:
            raise ValueError('Not enough data')

        return self.data[16:51]

    def set_input_report_id(self, _id):
        """
        :param _id: e.g. 0x21 Standard input reports used for sub command replies
                         0x30 Input reports with IMU data instead of sub command replies
                         etc... (TODO)
        """
        self.data[1] = _id

    def get_input_report_id(self):
        return self.data[1]

    def set_timer(self, timer):
        """
        Input report timer [0x00-0xFF], usually set by the transport
        """
        self.data[2] = timer % 256

    def set_misc(self):
        # battery level + connection info
        self.data[3] = 0x8E

    def set_button_status(self, button_status):
        """
        Sets the button status bytes
        """
        self.data[4:7] = iter(button_status)

    def set_stick_status(self, left_stick, right_stick):
        """
        Sets the joystick status bytes
        """
        self.set_left_analog_stick(bytes(left_stick))
        self.set_right_analog_stick(bytes(right_stick))

    def set_left_analog_stick(self, left_stick_bytes):
        """
        Set left analog stick status bytes.
        :param left_stick_bytes: 3 bytes
        """
        if len(left_stick_bytes) != 3:
            raise ValueError('Left stick status data must be exactly 3 bytes!')
        self.data[7:10] = left_stick_bytes

    def set_right_analog_stick(self, right_stick_bytes):
        """
        Set right analog stick status bytes.
        :param right_stick_bytes: 3 bytes
        """
        if len(right_stick_bytes) != 3:
            raise ValueError('Right stick status data must be exactly 3 bytes!')
        self.data[10:13] = right_stick_bytes

    def set_vibrator_input(self):
        """
        TODO
        """
        self.data[13] = 0x80

    def set_ack(self, ack):
        """
        ACK byte for subcmd reply
        TODO
        """
        self.data[14] = ack

    def get_ack(self):
        return self.data[14]

    def set_6axis_data(self):
        """
        Set accelerator and gyro of 0x30 input reports
        TODO
        """
        # HACK: Set all 0 for now
        for i in range(14, 50):
            self.data[i] = 0x00

    def set_ir_nfc_data(self, data):
        if 50 + len(data) > len(self.data):
            raise ValueError('Too much data.')

        # write to data
        for i in range(len(data)):
            self.data[50 + i] = data[i]

    def reply_to_subcommand_id(self, _id):
        if isinstance(_id, SubCommand):
            self.data[15] = _id.value
        else:
            self.data[15] = _id

    def get_reply_to_subcommand_id(self):
        if len(self.data) < 16:
            return None
        try:
            return SubCommand(self.data[15])
        except ValueError:
            raise NotImplementedError(f'Sub command id {hex(self.data[11])} not implemented')

    def sub_0x02_device_info(self, mac, fm_version=(0x04, 0x00), controller=Controller.JOYCON_L):
        """
        Sub command 0x02 request device info response.

        :param mac: Controller MAC address in Big Endian (6 Bytes)
        :param fm_version: TODO
        :param controller: 1=Left Joy-Con, 2=Right Joy-Con, 3=Pro Controller
        """
        if len(fm_version) != 2:
            raise ValueError('Firmware version must consist of 2 bytes!')
        elif len(mac) != 6:
            raise ValueError('Bluetooth mac address must consist of 6 bytes!')

        self.reply_to_subcommand_id(0x02)

        # sub command reply data
        offset = 16
        self.data[offset: offset + 2] = fm_version
        self.data[offset + 2] = controller.value
        self.data[offset + 3] = 0x02
        self.data[offset + 4: offset + 10] = mac
        self.data[offset + 10] = 0x01
        self.data[offset + 11] = 0x01

    def sub_0x10_spi_flash_read(self, offset, size, data):
        if len(data) != size:
            raise ValueError(f'Length of data {len(data)} does not match size {size}')
        if size > 0x1D:
            raise ValueError(f'Size can not exceed {0x1D}')

        self.reply_to_subcommand_id(0x10)

        # write offset to data
        for i in range(16, 16 + 4):
            self.data[i] = offset % 0x100
            offset = offset // 0x100

        self.data[20] = size
        self.data[21:21+len(data)] = data

    def sub_0x04_trigger_buttons_elapsed_time(self, L_ms=0, R_ms=0, ZL_ms=0, ZR_ms=0, SL_ms=0, SR_ms=0, HOME_ms=0):
        """
        Set sub command data for 0x04 reply. Arguments are in ms and must be divisible by 10.
        """
        if any(ms > 10*0xffff for ms in (L_ms, R_ms, ZL_ms, ZR_ms, SL_ms, SR_ms, HOME_ms)):
            raise ValueError(f'Values can not exceed {10*0xffff} ms.')

        def set(offset, ms):
            # reply data offset
            sub_command_offset = 16
            value = int(ms // 10)
            self.data[sub_command_offset + offset] = 0xff & value
            self.data[sub_command_offset + offset + 1] = (0xff00 & value) >> 8

        set(0, L_ms)
        set(2, R_ms)
        set(4, ZL_ms)
        set(6, ZR_ms)
        set(8, SL_ms)
        set(10, SR_ms)
        set(12, HOME_ms)

    def __bytes__(self):
        _id = self.get_input_report_id()
        if _id == 0x21:
            return bytes(self.data[:51])
        elif _id == 0x30:
            return bytes(self.data[:14])
        elif _id == 0x31:
            return bytes(self.data[:363])
        else:
            return bytes(self.data[:51])

    def __str__(self):
        _id = f'Input {self.get_input_report_id():x}'
        _info = ''
        if self.get_input_report_id() == 0x21:
            _info = self.get_reply_to_subcommand_id()
        _bytes = ' '.join(f'{byte:x}' for byte in bytes(self))

        return f'{_id} {_info}\n{_bytes}'


class SubCommand(Enum):
    REQUEST_DEVICE_INFO = 0x02
    SET_INPUT_REPORT_MODE = 0x03
    TRIGGER_BUTTONS_ELAPSED_TIME = 0x04
    SET_SHIPMENT_STATE = 0x08
    SPI_FLASH_READ = 0x10
    SET_NFC_IR_MCU_CONFIG = 0x21
    SET_NFC_IR_MCU_STATE = 0x22
    SET_PLAYER_LIGHTS = 0x30
    ENABLE_6AXIS_SENSOR = 0x40
    ENABLE_VIBRATION = 0x48


class OutputReportID(Enum):
    SUB_COMMAND = 0x01
    RUMBLE_ONLY = 0x10
    REQUEST_IR_NFC_MCU = 0x11


class OutputReport:
    def __init__(self, data=None):
        if not data:
            data = 50 * [0x00]
            data[0] = 0xA2
        elif data[0] != 0xA2:
            raise ValueError('Output reports must start with a 0xA2 byte!')
        self.data = data

    def get_output_report_id(self):
        try:
            return OutputReportID(self.data[1])
        except ValueError:
            raise NotImplementedError(f'Output report id {hex(self.data[1])} not implemented')

    def set_output_report_id(self, _id):
        if isinstance(_id, OutputReportID):
            self.data[1] = _id.value
        else:
            self.data[1] = _id

    def get_timer(self):
        return OutputReportID(self.data[2])

    def set_timer(self, timer):
        """
        Output report timer in [0x0, 0xF]
        """
        self.data[2] = timer % 0x10

    def get_rumble_data(self):
        return self.data[3:11]

    def get_sub_command(self):
        if len(self.data) < 12:
            return None
        try:
            return SubCommand(self.data[11])
        except ValueError:
            raise NotImplementedError(f'Sub command id {hex(self.data[11])} not implemented')

    def set_sub_command(self, _id):
        if isinstance(_id, SubCommand):
            self.data[11] = _id.value
        elif isinstance(_id, int):
            self.data[11] = _id
        else:
            raise ValueError('id must be int or SubCommand')

    def get_sub_command_data(self):
        if len(self.data) < 13:
            return None
        return self.data[12:]

    def set_sub_command_data(self, data):
        for i, _byte in enumerate(data):
            self.data[12+i] = _byte

    def sub_0x10_spi_flash_read(self, offset, size):
        """
        Creates output report data with spi flash read sub command.
        :param offset: start byte of the spi flash to read in [0x00, 0x80000)
        :param size: size of data to be read in [0x00, 0x1D]
        """
        if size > 0x1D:
            raise ValueError(f'Size read can not exceed {0x1D}')
        if offset+size > 0x80000:
            raise ValueError(f'Given address range exceeds max address {0x80000-1}')

        self.set_output_report_id(OutputReportID.SUB_COMMAND)
        self.set_sub_command(SubCommand.SPI_FLASH_READ)

        # write offset to data
        for i in range(12, 12+4):
            self.data[i] = offset % 0x100
            offset = offset // 0x100

        self.data[16] = size

    def __bytes__(self):
        return bytes(self.data)

    def __str__(self):
        _id = f'Output {self.get_output_report_id()}'
        _info = ''
        if self.get_output_report_id() == OutputReportID.SUB_COMMAND:
            _info = self.get_sub_command()
        _bytes = ' '.join(f'{byte:x}' for byte in bytes(self))

        return f'{_id} {_info}\n{_bytes}'



from enum import Enum
import os
import logging

import dbus


class ControllerTypes(Enum):
    """Controller type enumerations for initializing the controller server.
    """

    JOYCON_L = 1
    JOYCON_R = 2
    PRO_CONTROLLER = 3


class Controller():

    GAMEPAD_CLASS = "0x002508"
    SDP_UUID = "00001000-0000-1000-8000-00805f9b34fb"
    SDP_RECORD_PATH = "/nxbt/controller"
    ALIASES = {
        ControllerTypes.JOYCON_L: "Joy-Con (L)",
        ControllerTypes.JOYCON_R: "Joy-Con (R)",
        ControllerTypes.PRO_CONTROLLER: "Pro Controller"
    }

    def __init__(self, bluetooth, controller_type):

        self.bt = bluetooth
        self.logger = logging.getLogger('nxbt')

        if controller_type not in self.ALIASES.keys():
            raise ValueError("Unknown controller type specified")
        self.alias = self.ALIASES[controller_type]

    def setup(self):
        """Configures the specified Bluetooth device as the
        specified controller.
        """

        # Setting up Bluetooth adapter options
        self.bt.set_powered(True)
        self.bt.set_pairable(True)
        self.bt.set_pairable_timeout(0)
        self.bt.set_discoverable_timeout(180)

        self.bt.set_alias(self.alias)

        # Adding the SDP record
        sdp_record_path = os.path.join(
            os.path.dirname(__file__), "sdp", "switch-controller.xml")
        sdp_record = None
        with open(sdp_record_path, "r") as f:
            sdp_record = f.read()

        opts = {
            "ServiceRecord": sdp_record,
            "Role": "server",
            "RequireAuthentication": False,
            "RequireAuthorization": False,
            "AutoConnect": True
        }
        # If the profile has already been registered,
        # catch the error and continue
        try:
            self.bt.register_profile(self.SDP_RECORD_PATH, self.SDP_UUID, opts)
        except dbus.exceptions.DBusException as e:
            self.logger.debug(e)



import argparse
from random import randint
from time import sleep
import os
import traceback

from .nxbt import Nxbt, PRO_CONTROLLER
from .bluez import find_devices_by_alias
from .tui import InputTUI


parser = argparse.ArgumentParser()
parser.add_argument('command', default=False, choices=[
                        'webapp', 'demo', 'macro', 'tui', 'remote_tui', 'addresses', 'test'
                    ],
                    help="""Specifies the nxbt command to run:
                    webapp - Runs web server and allows for controller/macro
                    input from a web browser.
                    demo - Runs a demo macro (please ensure that your Switch
                    is on the main menu's Change Grip/Order menu before running).
                    macro - Allows for input of a specified macro from the command line
                    (with the argument -s) or from a file (with the argument -f).
                    tui/remote_tui - Opens a TUI that allows for direct input from the keyboard
                    to the Switch. 
                    addresses - Lists the Bluetooth MAC addresses for
                    all previously connected Nintendo Switches.
                    test - Runs through a series of tests to ensure NXBT is working and
                    compatible with your system.""")
parser.add_argument('-c', '--commands', required=False, default=False,
                    help="""Used in conjunction with the macro command. Specifies a
                    macro string or a file location to load a macro string from.""")
parser.add_argument('-r', '--reconnect', required=False, default=False, action='store_true',
                    help="""Used in conjunction with the macro or tui command. If specified,
                    nxbt will attmept to reconnect to any previously connected
                    Nintendo Switch.""")
parser.add_argument('-a', '--address', required=False, default=False,
                    help="""Used in conjunction with the macro or tui command. If specified,
                    nxbt will attmept to reconnect to a specific Bluetooth MAC address
                    of a Nintendo Switch.""")
parser.add_argument('-d', '--debug', required=False, default=False, action='store_true',
                    help="""Enables debug mode in nxbt.""")
parser.add_argument('-l', '--logfile', required=False, default=False, action='store_true',
                    help="""Enables logging to a file in the current working directory
                    instead of stderr.""")
parser.add_argument('-i', '--ip', required=False, default="0.0.0.0", type=str,
                    help="""Specifies the IP to run the webapp at. Defaults to 0.0.0.0""")
parser.add_argument('-p', '--port', required=False, default=8000, type=int,
                    help="""Specifies the port to run the webapp at. Defaults to 8000""")
parser.add_argument('--usessl', required=False, default=False, action='store_true',
                    help="""Enables or disables SSL use in the webapp""")
parser.add_argument('--certpath', required=False, default=None, type=str,
                    help="""Specifies the folder location for SSL certificates used
                    in the webapp. Certificates in this folder should be in the form of
                    a 'cert.pem' and 'key.pem' pair.""")                
args = parser.parse_args()


MACRO = """
B 0.1s
0.5s
B 0.1s
0.5s
B 0.1s
0.5s
B 0.1s
1.5s
DPAD_RIGHT 0.075s
0.075s
A 0.1s
1.5s
LOOP 12
    DPAD_DOWN 0.075s
    0.075s
A 0.1s
0.25s
DPAD_DOWN 0.93s
A 0.1s
0.25s
L_STICK_PRESS 0.1s
1.0s
L_STICK@-100+000 0.75s
L_STICK@+000+100 0.75s
L_STICK@+100+000 0.75s
L_STICK@+000-100 0.75s
B 0.1s
0.25s
R_STICK_PRESS 0.1s
1.0s
R_STICK@-100+000 0.75s
R_STICK@+000+100 0.75s
R_STICK@+100+000 0.75s
R_STICK@+000-100 0.75s
B 0.1s
0.1s
B 0.1s
0.1s
B 0.1s
0.1s
B 0.1s
0.4s
DPAD_LEFT 0.1s
0.1s
A 0.1s
1.5s
A 0.1s
5.0s
"""


def random_colour():

    return [
        randint(0, 255),
        randint(0, 255),
        randint(0, 255),
    ]


def check_bluetooth_address(address):
    """Check the validity of a given Bluetooth MAC address

    :param address: A Bluetooth MAC address
    :type address: str
    :raises ValueError: If the Bluetooth address is invalid
    """

    address_bytes = len(address.split(":"))
    if address_bytes != 6:
        raise ValueError("Invalid Bluetooth address")


def get_reconnect_target():

    if args.reconnect:
        reconnect_target = find_devices_by_alias("Nintendo Switch")
    elif args.address:
        check_bluetooth_address(args.address)
        reconnect_target = args.address
    else:
        reconnect_target = None

    return reconnect_target


def demo():
    """Loops over all available Bluetooth adapters
    and creates controllers on each. The last available adapter
    is used to run a macro.
    """

    nx = Nxbt(debug=args.debug, log_to_file=args.logfile)
    adapters = nx.get_available_adapters()
    if len(adapters) < 1:
        raise OSError("Unable to detect any Bluetooth adapters.")

    controller_idxs = []
    for i in range(0, len(adapters)):
        index = nx.create_controller(
            PRO_CONTROLLER,
            adapters[i],
            colour_body=random_colour(),
            colour_buttons=random_colour())
        controller_idxs.append(index)

    # Run a macro on the last controller
    print("Running Demo...")
    macro_id = nx.macro(controller_idxs[-1], MACRO, block=False)
    while macro_id not in nx.state[controller_idxs[-1]]["finished_macros"]:
        state = nx.state[controller_idxs[-1]]
        if state['state'] == 'crashed':
            print("An error occurred while running the demo:")
            print(state['errors'])
            exit(1)
        sleep(1.0)

    print("Finished!")


def test():
    """Tests NXBT functionality"""
    # Init
    print("[1] Attempting to initialize NXBT...")
    nx = None
    try:
        nx = Nxbt(debug=args.debug, log_to_file=args.logfile)
    except Exception as e:
        print("Failed to initialize:")
        print(traceback.format_exc())
        exit(1)
    print("Successfully initialized NXBT.\n")

    # Adapter Check
    print("[2] Checking for Bluetooth adapter availability...")
    adapters = None
    try:
        adapters = nx.get_available_adapters()
    except Exception as e:
        print("Failed to check for adapters:")
        print(traceback.format_exc())
        exit(1)
    if len(adapters) < 1:
        print("Unable to detect any Bluetooth adapters.")
        print("Please ensure you system has Bluetooth capability.")
        exit(1)
    print(f"{len(adapters)} Bluetooth adapter(s) available.")
    print("Adapters:", adapters, "\n")

    # Creating a controller
    print("[3] Please turn on your Switch and navigate to the 'Change Grip/Order menu.'")
    input("Press Enter to continue...")

    print("Creating a controller with the first Bluetooth adapter...")
    cindex = None
    try:
        cindex = nx.create_controller(
                 PRO_CONTROLLER,
                 adapters[0],
                 colour_body=random_colour(),
                 colour_buttons=random_colour())
    except Exception as e:
        print("Failed to create a controller:")
        print(traceback.format_exc())
        exit(1)
    print("Successfully created a controller.\n")

    # Controller connection check
    print("[4] Waiting for controller to connect with the Switch...")
    timeout = 120
    print(f"Connection timeout is {timeout} seconds for this test script.")
    elapsed = 0
    while nx.state[cindex]['state'] != 'connected':
        if elapsed >= timeout:
            print("Timeout reached, exiting...")
            exit(1)
        elif nx.state[cindex]['state'] == 'crashed':
            print("An error occurred while connecting:")
            print(nx.state[cindex]['errors'])
            exit(1)
        elapsed += 1
        sleep(1)
    print("Successfully connected.\n")

    # Exit the Change Grip/Order Menu
    print("[5] Attempting to exit the 'Change Grip/Order Menu'...")
    nx.macro(cindex, "B 0.1s\n0.1s")
    sleep(5)
    if nx.state[cindex]['state'] != 'connected':
        print("Controller disconnected after leaving the menu.")
        print("Exiting...")
        exit(1)
    print("Controller successfully exited the menu.\n")

    print("All tests passed.")


def macro():
    """Runs a macro from the command line.
    The macro can be from a specified file, a command line string,
    or input from the user in an interactive process.
    """

    macro = None
    if args.commands:
        if os.path.isfile(args.commands):
            with open(args.commands, "r") as f:
                macro = f.read()
        else:
            macro = args.commands
    else:
        print("No macro commands were specified.")
        print("Please use the -c argument to specify a macro string or a file location")
        print("to load a macro string from.")
        return

    reconnect_target = get_reconnect_target()

    nx = Nxbt(debug=args.debug, log_to_file=args.logfile)
    print("Creating controller...")
    index = nx.create_controller(
        PRO_CONTROLLER,
        colour_body=random_colour(),
        colour_buttons=random_colour(),
        reconnect_address=reconnect_target)
    print("Waiting for connection...")
    nx.wait_for_connection(index)
    print("Connected!")

    print("Running macro...")
    macro_id = nx.macro(index, macro, block=False)
    while (True):
        if nx.state[index]["state"] == "crashed":
            print("Controller crashed while running macro")
            print(nx.state[index]["errors"])
            break
        if macro_id in nx.state[index]["finished_macros"]:
            print("Finished running macro. Exiting...")
            break
        sleep(1/30)


def list_switch_addresses():

    addresses = find_devices_by_alias("Nintendo Switch")

    if not addresses or len(addresses) < 1:
        print("No Switches have previously connected to this device.")
        return

    print("---------------------------")
    print("| Num | Address           |")
    print("---------------------------")
    for i in range(0, len(addresses)):
        address = addresses[i]
        print(f"| {i+1}   | {address} |")
    print("---------------------------")


def main():

    if args.command == 'webapp':
        from .web import start_web_app
        start_web_app(ip=args.ip, port=args.port,
            usessl=args.usessl, cert_path=args.certpath)
    elif args.command == 'demo':
        demo()
    elif args.command == 'macro':
        macro()
    elif args.command == 'tui':
        reconnect_target = get_reconnect_target()
        tui = InputTUI(reconnect_target=reconnect_target)
        tui.start()
    elif args.command == 'remote_tui':
        reconnect_target = get_reconnect_target()
        tui = InputTUI(reconnect_target=reconnect_target, force_remote=True)
        tui.start()
    elif args.command == 'addresses':
        list_switch_addresses()
    elif args.command == 'test':
        test()



import asyncio
import logging
import time
from asyncio import BaseTransport, BaseProtocol
from contextlib import suppress
from typing import Optional, Union, Tuple, Text

from joycontrol import utils
from joycontrol.controller import Controller
from joycontrol.controller_state import ControllerState
from joycontrol.memory import FlashMemory
from joycontrol.report import OutputReport, SubCommand, InputReport, OutputReportID
from joycontrol.transport import NotConnectedError

logger = logging.getLogger(__name__)


def controller_protocol_factory(controller: Controller, spi_flash=None):
    if isinstance(spi_flash, bytes):
        spi_flash = FlashMemory(spi_flash_memory_data=spi_flash)

    def create_controller_protocol():
        return ControllerProtocol(controller, spi_flash=spi_flash)

    return create_controller_protocol


class ControllerProtocol(BaseProtocol):
    def __init__(self, controller: Controller, spi_flash: FlashMemory = None):
        self.controller = controller
        self.spi_flash = spi_flash

        self.transport = None

        # Increases for each input report send, should overflow at 0x100
        self._input_report_timer = 0x00

        self._data_received = asyncio.Event()

        self._controller_state = ControllerState(self, controller, spi_flash=spi_flash)
        self._controller_state_sender = None

        # None = Just answer to sub commands
        self._input_report_mode = None

        # This event gets triggered once the Switch assigns a player number to the controller and accepts user inputs
        self.sig_set_player_lights = asyncio.Event()

    async def send_controller_state(self):
        """
        Waits for the controller state to be send.

        Raises NotConnected exception if the transport is not connected or the connection was lost.
        """
        # TODO: Call write directly if in continuously sending input report mode

        if self.transport is None:
            raise NotConnectedError('Transport not registered.')

        self._controller_state.sig_is_send.clear()

        # wrap into a future to be able to set an exception in case of a disconnect
        self._controller_state_sender = asyncio.ensure_future(self._controller_state.sig_is_send.wait())
        await self._controller_state_sender
        self._controller_state_sender = None

    async def write(self, input_report: InputReport):
        """
        Sets timer byte and current button state in the input report and sends it.
        Fires sig_is_send event in the controller state afterwards.

        Raises NotConnected exception if the transport is not connected or the connection was lost.
        """
        if self.transport is None:
            raise NotConnectedError('Transport not registered.')

        # set button and stick data of input report
        input_report.set_button_status(self._controller_state.button_state)
        if self._controller_state.l_stick_state is None:
            l_stick = [0x00, 0x00, 0x00]
        else:
            l_stick = self._controller_state.l_stick_state
        if self._controller_state.r_stick_state is None:
            r_stick = [0x00, 0x00, 0x00]
        else:
            r_stick = self._controller_state.r_stick_state
        input_report.set_stick_status(l_stick, r_stick)

        # set timer byte of input report
        input_report.set_timer(self._input_report_timer)
        self._input_report_timer = (self._input_report_timer + 1) % 0x100

        await self.transport.write(input_report)

        self._controller_state.sig_is_send.set()

    def get_controller_state(self) -> ControllerState:
        return self._controller_state

    async def wait_for_output_report(self):
        """
        Waits until an output report from the Switch is received.
        """
        self._data_received.clear()
        await self._data_received.wait()

    def connection_made(self, transport: BaseTransport) -> None:
        logger.debug('Connection established.')
        self.transport = transport

    def connection_lost(self, exc: Optional[Exception] = None) -> None:
        if self.transport is not None:
            logger.error('Connection lost.')
            asyncio.ensure_future(self.transport.close())
            self.transport = None

            if self._controller_state_sender is not None:
                self._controller_state_sender.set_exception(NotConnectedError)

    def error_received(self, exc: Exception) -> None:
        # TODO?
        raise NotImplementedError()

    async def input_report_mode_full(self):
        """
        Continuously sends:
            0x30 input reports containing the controller state OR
            0x31 input reports containing the controller state and nfc data
        """
        if self.transport.is_reading():
            raise ValueError('Transport must be paused in full input report mode')

        # send state at 66Hz
        send_delay = 0.015
        await asyncio.sleep(send_delay)
        last_send_time = time.time()

        input_report = InputReport()
        input_report.set_vibrator_input()
        input_report.set_misc()
        if self._input_report_mode is None:
            raise ValueError('Input report mode is not set.')
        input_report.set_input_report_id(self._input_report_mode)

        reader = asyncio.ensure_future(self.transport.read())

        try:
            while True:
                reply_send = False
                if reader.done():
                    data = await reader

                    reader = asyncio.ensure_future(self.transport.read())

                    try:
                        report = OutputReport(list(data))
                        output_report_id = report.get_output_report_id()

                        if output_report_id == OutputReportID.RUMBLE_ONLY:
                            # TODO
                            pass
                        elif output_report_id == OutputReportID.SUB_COMMAND:
                            reply_send = await self._reply_to_sub_command(report)
                        elif output_report_id == OutputReportID.REQUEST_IR_NFC_MCU:
                            # TODO NFC
                            raise NotImplementedError('NFC communictation is not implemented.')                            
                        else:
                            logger.warning(f'Report unknown output report "{output_report_id}" - IGNORE')
                    except ValueError as v_err:
                        logger.warning(f'Report parsing error "{v_err}" - IGNORE')
                    except NotImplementedError as err:
                        logger.warning(err)

                if reply_send:
                    # Hack: Adding a delay here to avoid flooding during pairing
                    await asyncio.sleep(0.3)
                else:
                    # write 0x30 input report.
                    # TODO: set some sensor data
                    input_report.set_6axis_data()

                    # TODO NFC - set nfc data
                    if input_report.get_input_report_id() == 0x31:
                        pass

                    await self.write(input_report)

                # calculate delay
                current_time = time.time()
                time_delta = time.time() - last_send_time
                sleep_time = send_delay - time_delta
                last_send_time = current_time

                if sleep_time < 0:
                    # logger.warning(f'Code is running {abs(sleep_time)} s too slow!')
                    sleep_time = 0

                await asyncio.sleep(sleep_time)

        except NotConnectedError as err:
            # Stop 0x30 input report mode if disconnected.
            logger.error(err)
        finally:
            # cleanup
            self._input_report_mode = None
            # cancel the reader
            with suppress(asyncio.CancelledError, NotConnectedError):
                if reader.cancel():
                    await reader

    async def report_received(self, data: Union[bytes, Text], addr: Tuple[str, int]) -> None:
        self._data_received.set()

        try:
            report = OutputReport(list(data))
        except ValueError as v_err:
            logger.warning(f'Report parsing error "{v_err}" - IGNORE')
            return

        try:
            output_report_id = report.get_output_report_id()
        except NotImplementedError as err:
            logger.warning(err)
            return

        if output_report_id == OutputReportID.SUB_COMMAND:
            await self._reply_to_sub_command(report)
        # elif output_report_id == OutputReportID.RUMBLE_ONLY:
        #    pass
        else:
            logger.warning(f'Output report {output_report_id} not implemented - ignoring')

    async def _reply_to_sub_command(self, report):
        # classify sub command
        try:
            sub_command = report.get_sub_command()
        except NotImplementedError as err:
            logger.warning(err)
            return False

        if sub_command is None:
            raise ValueError('Received output report does not contain a sub command')

        logging.info(f'received output report - Sub command {sub_command}')

        sub_command_data = report.get_sub_command_data()
        assert sub_command_data is not None

        try:
            # answer to sub command
            if sub_command == SubCommand.REQUEST_DEVICE_INFO:
                await self._command_request_device_info(sub_command_data)

            elif sub_command == SubCommand.SET_SHIPMENT_STATE:
                await self._command_set_shipment_state(sub_command_data)

            elif sub_command == SubCommand.SPI_FLASH_READ:
                await self._command_spi_flash_read(sub_command_data)

            elif sub_command == SubCommand.SET_INPUT_REPORT_MODE:
                await self._command_set_input_report_mode(sub_command_data)

            elif sub_command == SubCommand.TRIGGER_BUTTONS_ELAPSED_TIME:
                await self._command_trigger_buttons_elapsed_time(sub_command_data)

            elif sub_command == SubCommand.ENABLE_6AXIS_SENSOR:
                await self._command_enable_6axis_sensor(sub_command_data)

            elif sub_command == SubCommand.ENABLE_VIBRATION:
                await self._command_enable_vibration(sub_command_data)

            elif sub_command == SubCommand.SET_NFC_IR_MCU_CONFIG:
                await self._command_set_nfc_ir_mcu_config(sub_command_data)

            elif sub_command == SubCommand.SET_NFC_IR_MCU_STATE:
                await self._command_set_nfc_ir_mcu_state(sub_command_data)

            elif sub_command == SubCommand.SET_PLAYER_LIGHTS:
                await self._command_set_player_lights(sub_command_data)
            else:
                logger.warning(f'Sub command 0x{sub_command.value:02x} not implemented - ignoring')
                return False
        except NotImplementedError as err:
            logger.error(f'Failed to answer {sub_command} - {err}')
            return False
        return True

    async def _command_request_device_info(self, sub_command_data):
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        address = self.transport.get_extra_info('sockname')
        assert address is not None
        bd_address = list(map(lambda x: int(x, 16), address[0].split(':')))

        input_report.set_ack(0x82)
        input_report.sub_0x02_device_info(bd_address, controller=self.controller)

        await self.write(input_report)

    async def _command_set_shipment_state(self, sub_command_data):
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(0x08)

        await self.write(input_report)

    async def _command_spi_flash_read(self, sub_command_data):
        """
        Replies with 0x21 input report containing requested data from the flash memory.
        :param sub_command_data: input report sub command data bytes
        """
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        input_report.set_ack(0x90)

        # parse offset
        offset = 0
        digit = 1
        for i in range(4):
            offset += sub_command_data[i] * digit
            digit *= 0x100

        size = sub_command_data[4]

        if self.spi_flash is not None:
            spi_flash_data = self.spi_flash[offset: offset + size]
            input_report.sub_0x10_spi_flash_read(offset, size, spi_flash_data)
        else:
            spi_flash_data = size * [0x00]
            input_report.sub_0x10_spi_flash_read(offset, size, spi_flash_data)

        await self.write(input_report)

    async def _command_set_input_report_mode(self, sub_command_data):
        if self._input_report_mode == sub_command_data[0]:
            logger.warning(f'Already in input report mode {sub_command_data[0]} - ignoring request')

        # Start input report reader
        if sub_command_data[0] in (0x30, 0x31):
            new_reader = asyncio.ensure_future(self.input_report_mode_full())
        else:
            logger.error(f'input report mode {sub_command_data[0]} not implemented - ignoring request')
            return

        # Replace the currently running reader with the input report mode sender,
        # which will also handle incoming requests in the future

        self.transport.pause_reading()

        # We need to replace the reader in the future because this function was probably called by it
        async def set_reader():
            await self.transport.set_reader(new_reader)

            logger.info(f'Setting input report mode to {hex(sub_command_data[0])}...')
            self._input_report_mode = sub_command_data[0]

            self.transport.resume_reading()

        asyncio.ensure_future(set_reader()).add_done_callback(
            utils.create_error_check_callback()
        )

        # Send acknowledgement
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(0x03)

        await self.write(input_report)

    async def _command_trigger_buttons_elapsed_time(self, sub_command_data):
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        input_report.set_ack(0x83)
        input_report.reply_to_subcommand_id(SubCommand.TRIGGER_BUTTONS_ELAPSED_TIME)
        # Hack: We assume this command is only used during pairing - Set values so the Switch assigns a player number
        if self.controller == Controller.PRO_CONTROLLER:
            input_report.sub_0x04_trigger_buttons_elapsed_time(L_ms=3000, R_ms=3000)
        elif self.controller in (Controller.JOYCON_L, Controller.JOYCON_R):
            # TODO: What do we do if we want to pair a combined JoyCon?
            input_report.sub_0x04_trigger_buttons_elapsed_time(SL_ms=3000, SR_ms=3000)
        else:
            raise NotImplementedError(self.controller)

        await self.write(input_report)

    async def _command_enable_6axis_sensor(self, sub_command_data):
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(0x40)

        await self.write(input_report)

    async def _command_enable_vibration(self, sub_command_data):
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(SubCommand.ENABLE_VIBRATION.value)

        await self.write(input_report)

    async def _command_set_nfc_ir_mcu_config(self, sub_command_data):
        # TODO NFC
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        input_report.set_ack(0xA0)
        input_report.reply_to_subcommand_id(SubCommand.SET_NFC_IR_MCU_CONFIG.value)

        data = [1, 0, 255, 0, 8, 0, 27, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 200]
        for i in range(len(data)):
            input_report.data[16 + i] = data[i]

        await self.write(input_report)

    async def _command_set_nfc_ir_mcu_state(self, sub_command_data):
        # TODO NFC
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        if sub_command_data[0] == 0x01:
            # 0x01 = Resume
            input_report.set_ack(0x80)
            input_report.reply_to_subcommand_id(SubCommand.SET_NFC_IR_MCU_STATE.value)
        elif sub_command_data[0] == 0x00:
            # 0x00 = Suspend
            input_report.set_ack(0x80)
            input_report.reply_to_subcommand_id(SubCommand.SET_NFC_IR_MCU_STATE.value)
        else:
            raise NotImplementedError(f'Argument {sub_command_data[0]} of {SubCommand.SET_NFC_IR_MCU_STATE} '
                                      f'not implemented.')

        await self.write(input_report)

    async def _command_set_player_lights(self, sub_command_data):
        input_report = InputReport()
        input_report.set_input_report_id(0x21)
        input_report.set_misc()

        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(SubCommand.SET_PLAYER_LIGHTS.value)

        await self.write(input_report)

        self.sig_set_player_lights.set()



import logging
import uuid
import dbus

from joycontrol import utils

logger = logging.getLogger(__name__)


HID_UUID = '00001124-0000-1000-8000-00805f9b34fb'
HID_PATH = '/bluez/switch/hid'


class HidDevice:
    def __init__(self, device_id=None):
        bus = dbus.SystemBus()

        # Get Bluetooth adapter from dbus interface
        manager = dbus.Interface(bus.get_object('org.bluez', '/'), 'org.freedesktop.DBus.ObjectManager')
        for path, ifaces in manager.GetManagedObjects().items():
            adapter_info = ifaces.get('org.bluez.Adapter1')
            if adapter_info is None:
                continue
            elif device_id is None or device_id == adapter_info['Address'] or path.endswith(str(device_id)):
                obj = bus.get_object('org.bluez', path)
                self.adapter = dbus.Interface(obj, 'org.bluez.Adapter1')
                self.address = adapter_info['Address']
                self._adapter_name = path.split('/')[-1]

                self.properties = dbus.Interface(self.adapter, 'org.freedesktop.DBus.Properties')
                break
        else:
            raise ValueError(f'Adapter {device_id} not found.')

    def get_address(self) -> str:
        """
        :returns adapter Bluetooth address
        """
        return self.address

    def powered(self, boolean=True):
        self.properties.Set(self.adapter.dbus_interface, 'Powered', boolean)

    def discoverable(self, boolean=True):
        """
        Make adapter discoverable, starts advertising.
        """
        self.properties.Set(self.adapter.dbus_interface, 'Discoverable', boolean)

    def pairable(self, boolean=True):
        """
        Make adapter pairable
        """
        self.properties.Set(self.adapter.dbus_interface, 'Pairable', boolean)

    async def set_class(self, cls='0x002508'):
        """
        Sets Bluetooth device class. Requires hciconfig system command.
        :param cls: default 0x002508 (Gamepad/joystick device class)
        """
        logger.info(f'setting device class to {cls}...')
        await utils.run_system_command(f'hciconfig {self._adapter_name} class {cls}')

    async def set_name(self, name: str):
        """
        Set Bluetooth device name.
        :param name: to set.
        """
        logger.info(f'setting device name to {name}...')
        self.properties.Set(self.adapter.dbus_interface, 'Alias', name)

    @staticmethod
    def register_sdp_record(record_path):
        _uuid = str(uuid.uuid4())

        with open(record_path) as record:
            opts = {
                'ServiceRecord': record.read(),
                'Role': 'server',
                'Service': HID_UUID,
                'RequireAuthentication': False,
                'RequireAuthorization': False
            }
            bus = dbus.SystemBus()
            manager = dbus.Interface(bus.get_object("org.bluez", "/org/bluez"), "org.bluez.ProfileManager1")
            manager.RegisterProfile(HID_PATH, _uuid, opts)

        return _uuid



from enum import Enum
import random
from time import perf_counter

from .controller import ControllerTypes
from .utils import replace_subarray


class SwitchResponses(Enum):

    NO_DATA = -1
    MALFORMED = -2
    TOO_SHORT = -3
    UNKNOWN_SUBCOMMAND = -4
    REQUEST_DEVICE_INFO = 2
    SET_SHIPMENT = 0x08
    SPI_READ = 0x10
    SET_MODE = 0x03
    TRIGGER_BUTTONS = 0x04
    TOGGLE_IMU = 0x40
    ENABLE_VIBRATION = 0x48
    SET_PLAYER = 0x30
    SET_NFC_IR_STATE = 0x22
    SET_NFC_IR_CONFIG = 0x21


class ControllerProtocol():

    CONTROLLER_INFO = {
        ControllerTypes.JOYCON_L: {
            "id": 0x01,
            "connection_info": 0x0E
        },
        ControllerTypes.JOYCON_R: {
            "id": 0x02,
            "connection_info": 0x0E
        },
        ControllerTypes.PRO_CONTROLLER: {
            "id": 0x03,
            "connection_info": 0x00
        }
    }
    VIBRATOR_BYTES = [0xA0, 0xB0, 0xC0, 0x90]

    def __init__(self, controller_type, bt_address, report_size=50,
                 colour_body=None, colour_buttons=None):
        """Initializes the protocol for the controller.

        :param controller_type: The type of controller (Joy-Con (L),
        Pro Controller, etc)
        :type controller_type: ControllerTypes
        :param bt_address: A colon-separated Bluetooth MAC address
        :type bt_address: string
        :param report_size: The size of the protocol report, defaults to 50
        :type report_size: int, optional
        :param colour_body: Sets the body colour of the controller, defaults
        to None
        :type colour_body: list of bytes, optional
        :param colour_buttons: Sets the colour of the controller buttons,
        defaults to None
        :type colour_buttons: list of bytes, optional
        :raises ValueError: On unknown controller type
        """

        self.bt_address = bt_address

        if controller_type in self.CONTROLLER_INFO.keys():
            self.controller_type = controller_type
        else:
            raise ValueError("Unknown controller type specified")

        self.report = None
        self.report_size = report_size
        self.set_empty_report()

        # Input report mode
        self.mode = None

        # Player number
        self.player_number = None

        # Setting if the controller has been asked for device info
        # Enables buttons/stick output for the standard full report
        self.device_info_queried = False

        # Standard Input Report Properties
        # Timestamp to generate timer byte ticks
        self.timer = 0
        self.timestamp = None

        # High/Low Nibble
        self.battery_level = 0x90
        self.connection_info = (
            self.CONTROLLER_INFO[self.controller_type]["connection_info"])

        self.button_status = [0x00] * 3

        # Disable left stick if we have a right Joy-Con
        if self.controller_type == ControllerTypes.JOYCON_R:
            self.left_stick_centre = [0x00] * 3
        else:
            # Center values which are also reported under
            # SPI Stick calibration reads
            self.left_stick_centre = [0x6F, 0xC8, 0x77]

        # Disable right stick if we have a left Joy-Con
        if self.controller_type == ControllerTypes.JOYCON_L:
            self.right_stick_centre = [0x00] * 3
        else:
            # Center values which are also reported under
            # SPI Stick calibration reads
            self.right_stick_centre = [0x16, 0xD8, 0x7D]

        self.vibration_enabled = False
        self.vibrator_report = random.choice(self.VIBRATOR_BYTES)

        # IMU (Six Axis Sensor) State
        self.imu_enabled = False

        # Controller colours
        # Body Colour
        if not colour_body:
            self.colour_body = [0x82] * 3
        else:
            self.colour_body = colour_body
        if not colour_buttons:
            self.colour_buttons = [0x0F] * 3
        else:
            self.colour_buttons = colour_buttons

    def get_report(self):

        report = bytes(self.report)
        # Clear report
        self.set_empty_report()
        return report

    def process_commands(self, data):

        # Parsing the Switch's message
        message = SwitchReportParser(data)

        # Responding to the parsed message
        if message.response == SwitchResponses.REQUEST_DEVICE_INFO:
            self.device_info_queried = True
            self.set_subcommand_reply()
            self.set_device_info()

        elif message.response == SwitchResponses.SET_SHIPMENT:
            self.set_subcommand_reply()
            self.set_shipment()

        elif message.response == SwitchResponses.SPI_READ:
            self.set_subcommand_reply()
            self.spi_read(message)

        elif message.response == SwitchResponses.SET_MODE:
            self.set_subcommand_reply()
            self.set_mode(message)

        elif message.response == SwitchResponses.TRIGGER_BUTTONS:
            self.set_subcommand_reply()
            self.set_trigger_buttons()

        elif message.response == SwitchResponses.TOGGLE_IMU:
            self.set_subcommand_reply()
            self.toggle_imu(message)

        elif message.response == SwitchResponses.ENABLE_VIBRATION:
            self.set_subcommand_reply()
            self.enable_vibration()

        elif message.response == SwitchResponses.SET_PLAYER:
            self.set_subcommand_reply()
            self.set_player_lights(message)

        elif message.response == SwitchResponses.SET_NFC_IR_STATE:
            self.set_subcommand_reply()
            self.set_nfc_ir_state()

        elif message.response == SwitchResponses.SET_NFC_IR_CONFIG:
            self.set_subcommand_reply()
            self.set_nfc_ir_config()

        # Bad Packet handling statements
        elif message.response == SwitchResponses.UNKNOWN_SUBCOMMAND:
            # Currently set so that the controller ignores any unknown
            # subcommands. This is better than sending a NACK response
            # since we'd just get stuck in an infinite loop arguing
            # with the Switch.
            self.set_full_input_report()

        elif message.response == SwitchResponses.NO_DATA:
            self.set_full_input_report()

        elif message.response == SwitchResponses.TOO_SHORT:
            self.set_full_input_report()

        elif message.response == SwitchResponses.MALFORMED:
            self.set_full_input_report()

    def set_empty_report(self):

        empty_report = [0] * self.report_size
        empty_report[0] = 0xA1

        self.report = empty_report

    def set_subcommand_reply(self):

        # Input Report ID
        self.report[1] = 0x21

        # TODO: Find out what the vibrator byte is doing.
        # This is a hack in an attempt to semi-emulate
        # actions of the vibrator byte as it seems to change
        # when a subcommand reply is sent.
        self.vibrator_report = random.choice(self.VIBRATOR_BYTES)

        self.set_standard_input_report()

    def set_unknown_subcommand(self, subcommand_id):

        # Set NACK
        self.report[14]

        # Set unknown subcommand ID
        self.report[15] = subcommand_id

    def set_timer(self):

        # If the timer hasn't been set before
        if not self.timestamp:
            self.timestamp = perf_counter()
            self.report[2] = 0x00
            return

        # Get the time that has passed since the last timestamp
        # in milliseconds
        now = perf_counter()
        delta_t = (now - self.timestamp) * 1000

        # Get how many ticks have passed in hex with overflow at 255
        # Joy-Con uses 4.96ms as the timer tick rate
        elapsed_ticks = int(delta_t * 4)
        self.timer = (self.timer + elapsed_ticks) & 0xFF

        self.report[2] = self.timer
        self.timestamp = now

    def set_full_input_report(self):

        # Setting Report ID to full standard input report ID
        self.report[1] = 0x30
        self.set_standard_input_report()
        self.set_imu_data()

    def set_standard_input_report(self):

        self.set_timer()

        if self.device_info_queried:
            self.report[3] = self.battery_level + self.connection_info

            self.report[4] = self.button_status[0]
            self.report[5] = self.button_status[1]
            self.report[6] = self.button_status[2]

            self.report[7] = self.left_stick_centre[0]
            self.report[8] = self.left_stick_centre[1]
            self.report[9] = self.left_stick_centre[2]

            self.report[10] = self.right_stick_centre[0]
            self.report[11] = self.right_stick_centre[1]
            self.report[12] = self.right_stick_centre[2]

            self.report[13] = self.vibrator_report

    def set_button_inputs(self, upper, shared, lower):

        self.report[4] = upper
        self.report[5] = shared
        self.report[6] = lower

    def set_left_stick_inputs(self, left):

        self.report[7] = left[0]
        self.report[8] = left[1]
        self.report[9] = left[2]

    def set_right_stick_inputs(self, right):

        self.report[10] = right[0]
        self.report[11] = right[1]
        self.report[12] = right[2]

    def set_device_info(self):

        # ACK Reply
        self.report[14] = 0x82

        # Subcommand Reply
        self.report[15] = 0x02

        # Firmware version
        self.report[16] = 0x03
        self.report[17] = 0x8B

        # Controller ID
        self.report[18] = self.CONTROLLER_INFO[self.controller_type]["id"]

        # Unknown Byte, always 2
        self.report[19] = 0x02

        # Controller Bluetooth Address
        address = self.bt_address.strip().split(":")  # Getting from adapter
        address_location = 20
        for address_byte_str in address:
            # Converting string address bytes to hex
            # and assigning to report
            address_byte = int(address_byte_str, 16)
            self.report[address_location] = address_byte
            address_location += 1

        # Unknown byte, always 1
        self.report[26] = 0x01

        # Controller colours location (read from SPI)
        self.report[27] = 0x01

    def set_shipment(self):

        # ACK Reply
        self.report[14] = 0x80

        # Subcommand reply
        self.report[15] = 0x08

    def toggle_imu(self, message):

        if message.subcommand[1] == 0x01:
            self.imu_enabled = True
        else:
            self.imu_enabled = False

        # ACK Reply
        self.report[14] = 0x80

        # Subcommand reply
        self.report[15] = 0x40

    def set_imu_data(self):

        if not self.imu_enabled:
            return

        imu_data = [0x75, 0xFD, 0xFD, 0xFF, 0x09, 0x10, 0x21, 0x00, 0xD5, 0xFF,
                    0xE0, 0xFF, 0x72, 0xFD, 0xF9, 0xFF, 0x0A, 0x10, 0x22, 0x00,
                    0xD5, 0xFF, 0xE0, 0xFF, 0x76, 0xFD, 0xFC, 0xFF, 0x09, 0x10,
                    0x23, 0x00, 0xD5, 0xFF, 0xE0, 0xFF]
        replace_subarray(self.report, 14, 49, replace_arr=imu_data)

    def spi_read(self, message):

        addr_top = message.subcommand[2]
        addr_bottom = message.subcommand[1]
        read_length = message.subcommand[5]

        # ACK byte
        self.report[14] = 0x90

        # Subcommand reply
        self.report[15] = 0x10

        # Read address
        self.report[16] = addr_bottom
        self.report[17] = addr_top

        # Read length
        self.report[20] = read_length

        # Stick Parameters
        # Params are generally the same for all sticks
        # Notable difference is the deadzone (10% Joy-Con vs 15% Pro Con)
        params = [0x0F, 0x30, 0x61,  # Unused
                  0x96, 0x30, 0xF3,  # Dead Zone/Range Ratio
                  0xD4, 0x14, 0x54,  # X/Y ?
                  0x41, 0x15, 0x54,  # X/Y ?
                  0xC7, 0x79, 0x9C,  # X/Y ?
                  0x33, 0x36, 0x63]  # X/Y ?
        # Adjusting deadzone for Joy-Cons
        if not self.controller_type == ControllerTypes.PRO_CONTROLLER:
            params[3] = 0xAE

        # Serial Number read
        if addr_top == 0x60 and addr_bottom == 0x00:
            # Switch will take this as no serial number
            replace_subarray(self.report, 21, 16, 0xFF)

        # Colours
        elif addr_top == 0x60 and addr_bottom == 0x50:
            # Body colour
            replace_subarray(
                self.report, 21, 3,
                replace_arr=self.colour_body)
            # Buttons colour
            replace_subarray(
                self.report, 24, 3,
                replace_arr=self.colour_buttons)
            # Left/right grip colours (Pro controller)
            replace_subarray(self.report, 27, 7, 0xFF)

        # Factory sensor/stick device parameters
        elif addr_top == 0x60 and addr_bottom == 0x80:

            # Six-Axis factory parameters
            if self.controller_type == ControllerTypes.PRO_CONTROLLER:
                self.report[21] = 0x50
                self.report[22] = 0xFD
                self.report[23] = 0x00
                self.report[24] = 0x00
                self.report[25] = 0xC6
                self.report[26] = 0x0F
            else:
                self.report[21] = 0x5E
                self.report[22] = 0x01
                self.report[23] = 0x00
                self.report[24] = 0x00
                if self.controller_type == ControllerTypes.JOYCON_L:
                    self.report[25] = 0xF1
                    self.report[26] = 0x0F
                else:
                    self.report[25] = 0x0F
                    self.report[26] = 0xF0

            replace_subarray(self.report, 27, 18, replace_arr=params)

        # Stick device parameters 2
        elif addr_top == 0x60 and addr_bottom == 0x98:

            # Setting same params since controllers always
            # have duplicates of stick params 1 for stick params 2
            replace_subarray(self.report, 21, 18, replace_arr=params)

        # User analog stick calibration
        elif addr_top == 0x80 and addr_bottom == 0x10:

            # Fill report with null user calibration info
            replace_subarray(self.report, 21, 24, 0xFF)

        # Factory analog stick calibration
        elif addr_top == 0x60 and addr_bottom == 0x3D:

            # Left/right stick calibration
            l_calibration = [0xBA, 0xF5, 0x62,
                             0x6F, 0xC8, 0x77,
                             0xED, 0x95, 0x5B]
            r_calibration = [0x16, 0xD8, 0x7D,
                             0xF2, 0xB5, 0x5F,
                             0x86, 0x65, 0x5E]

            # Left stick calibration
            # If null, fill with 0xFF
            if not self.controller_type == ControllerTypes.JOYCON_R:
                replace_subarray(self.report, 21, 9, replace_arr=l_calibration)
            else:
                replace_subarray(self.report, 21, 9, value=0xFF)

            # Right stick calibration
            # If null, fill with 0xFF
            if not self.controller_type == ControllerTypes.JOYCON_L:
                replace_subarray(self.report, 30, 9, replace_arr=r_calibration)
            else:
                replace_subarray(self.report, 30, 9, value=0xFF)

            # Spacer byte
            self.report[39] = 0xFF

            # Body colour
            replace_subarray(
                self.report, 40, 3,
                replace_arr=self.colour_body)
            # Buttons colour
            replace_subarray(
                self.report, 43, 3,
                replace_arr=self.colour_buttons)

        # Six-Axis motion sensor factor calibration
        elif addr_top == 0x60 and addr_bottom == 0x20:

            # 1: Acceleration origin position
            # 2: Acceleration sensitivity coefficient
            # 3: Gyro origin when still
            # 4: Gyro sensitivity coefficient
            sa_calibration = [0xD3, 0xFF, 0xD5, 0xFF, 0x55, 0x01,  # 1
                              0x00, 0x40, 0x00, 0x40, 0x00, 0x40,  # 2
                              0x19, 0x00, 0xDD, 0xFF, 0xDC, 0xFF,  # 3
                              0x3B, 0x34, 0x3B, 0x34, 0x3B, 0x34]  # 4

            replace_subarray(self.report, 21, 24, replace_arr=sa_calibration)

    def set_mode(self, message):

        # ACK byte
        self.report[14] = 0x80

        # Subcommand reply
        self.report[15] = 0x03

        if message.subcommand[1] == 0x30:
            self.mode = "standard"
        elif message.subcommand[1] == 0x31:
            self.mode = "nfc/ir"
        elif message.subcommand[1] == 0x3F:
            self.mode = "simpleHID"

    def set_trigger_buttons(self):

        # ACK byte
        self.report[14] = 0x83

        # Subcommand reply
        self.report[15] = 0x04

    def enable_vibration(self):

        # ACK Reply
        self.report[14] = 0x82

        # Subcommand reply
        self.report[15] = 0x48

        # Set class property
        self.vibration_enabled = True

    def set_player_lights(self, message):

        # ACK byte
        self.report[14] = 0x80

        # Subcommand reply
        self.report[15] = 0x30

        bitfield = message.subcommand[1]

        if bitfield == 0x01 or bitfield == 0x10:
            self.player_number = 1
        elif bitfield == 0x03 or bitfield == 0x30:
            self.player_number = 2
        elif bitfield == 0x07 or bitfield == 0x70:
            self.player_number = 3
        elif bitfield == 0x0F or bitfield == 0xF0:
            self.player_number = 4

    def set_nfc_ir_state(self):

        # ACK byte
        self.report[14] = 0x80

        # Subcommand reply
        self.report[15] = 0x22

    def set_nfc_ir_config(self):

        # ACK byte
        self.report[14] = 0xA0

        # Subcommand reply
        self.report[15] = 0x21

        # NFC/IR state data
        params = [0x01, 0x00, 0xFF, 0x00, 0x08, 0x00, 0x1B, 0x01]
        replace_subarray(self.report, 16, 8, replace_arr=params)
        self.report[49] = 0xC8


class SwitchReportParser():

    SUBCOMMANDS = {
        0x02: SwitchResponses.REQUEST_DEVICE_INFO,
        0x08: SwitchResponses.SET_SHIPMENT,
        0x10: SwitchResponses.SPI_READ,
        0x03: SwitchResponses.SET_MODE,
        0x04: SwitchResponses.TRIGGER_BUTTONS,
        0x40: SwitchResponses.TOGGLE_IMU,
        0x48: SwitchResponses.ENABLE_VIBRATION,
        0x30: SwitchResponses.SET_PLAYER,
        0x22: SwitchResponses.SET_NFC_IR_STATE,
        0x21: SwitchResponses.SET_NFC_IR_CONFIG,
    }

    def __init__(self, data, data_length=50):

        # Non-data check
        if not data:
            self.response = SwitchResponses.NO_DATA
            return

        # Report length check
        if len(data) < data_length:
            self.response = SwitchResponses.TOO_SHORT
            return

        # First byte check
        if data[0] != 0xA2:
            self.response = SwitchResponses.MALFORMED
            return

        # Splitting data
        self.payload = data[:11]
        self.subcommand = data[11:]
        self.subcommand_id = self.subcommand[0]

        # Parsing the subcommand
        if self.subcommand[0] in self.SUBCOMMANDS.keys():
            self.response = self.SUBCOMMANDS[self.subcommand[0]]
        else:
            self.response = SwitchResponses.UNKNOWN_SUBCOMMAND



import asyncio
import logging
from contextlib import contextmanager

import hid

logger = logging.getLogger(__name__)


class AsyncHID(hid.Device):
    def __init__(self, *args, loop=asyncio.get_event_loop(), **kwargs):
        super().__init__(*args, **kwargs)
        self._loop = loop

        self._write_lock = asyncio.Lock()
        self._read_lock = asyncio.Lock()

    async def read(self, size, timeout=None):
        async with self._read_lock:
            return await self._loop.run_in_executor(None, hid.Device.read, self, size, timeout)

    async def write(self, data):
        async with self._write_lock:
            return await self._loop.run_in_executor(None, hid.Device.write, self, data)


@contextmanager
def get_output(path=None, open_flags='wb', default=None):
    """
    Context manager that open the file a path was given, otherwise returns default value.
    """
    if path is not None:
        file = open(path, open_flags)
        yield file
        file.close()
    else:
        yield default


def get_bit(value, n):
    return (value >> n & 1) != 0


def flip_bit(value, n):
    return value ^ (1 << n)


def create_error_check_callback(ignore=None):
    """
    Creates callback causing errors of a finished future to be raised.
    Useful for debugging futures that are never awaited.
    :param ignore: Any number of errors to ignore.
    :returns callback which can be added to a future with future.add_done_callback(...)
    """
    def callback(future):
        if ignore:
            try:
                future.result()
            except ignore:
                # ignore suppressed errors
                pass
        else:
            future.result()
    return callback


async def run_system_command(cmd):
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)

    stdout, stderr = await proc.communicate()

    logger.debug(f'[{cmd!r} exited with {proc.returncode}]')
    if stdout:
        logger.debug(f'[stdout]\n{stdout.decode()}')
    if stderr:
        logger.debug(f'[stderr]\n{stderr.decode()}')

    return proc.returncode, stdout, stderr

"""
async def get_bt_mac_address(dev=0):
    ret, stdout, stderr = await run_system_command(f'hciconfig hci{dev}')
    # TODO: Process error handling

    match = re.search(r'BD Address: (?P<mac>\w\w:\w\w:\w\w:\w\w:\w\w:\w\w)', stdout.decode('UTF-8'))
    if match:
        return list(map(lambda x: int(x, 16), match.group('mac').split(':')))
    else:
        raise ValueError(f'BD Address not found in "{stdout}"')
"""



import os
import time
import psutil
from collections import deque
import multiprocessing

from blessed import Terminal

from .nxbt import Nxbt, PRO_CONTROLLER


class LoadingSpinner():

    SPINNER_CHARS = ['■ □ □ □', '□ ■ □ □', '□ □ ■ □', '□ □ □ ■', '□ □ □ ■', '□ □ ■ □', '□ ■ □ □', '■ □ □ □']  # noqa

    def __init__(self):

        self.creation_time = time.perf_counter()
        self.last_update_time = self.creation_time
        self.current_char_index = 0

    def get_spinner_char(self):

        current_time = time.perf_counter()
        delta = current_time - self.last_update_time

        if delta > 0.07:
            self.last_update_time = current_time

            if self.current_char_index == 7:
                self.current_char_index = 0
            else:
                self.current_char_index += 1

        return self.SPINNER_CHARS[self.current_char_index]


class ControllerTUI():

    CONTROLS = {
        "ZL": "◿□□□□",
        "L": "◿□□□□",
        "ZR": "□□□□◺",
        "R": "□□□□◺",
        "LS_UP": ".─.",
        "LS_LEFT": "(",
        "LS_RIGHT": ")",
        "LS_DOWN": "`─'",
        "RS_UP": ".─.",
        "RS_LEFT": "(",
        "RS_RIGHT": ")",
        "RS_DOWN": "`─'",
        "DPAD_UP": "△",
        "DPAD_LEFT": "◁",
        "DPAD_RIGHT": "▷",
        "DPAD_DOWN": "▽",
        "MINUS": "◎",
        "PLUS": "◎",
        "HOME": "□",
        "CAPTURE": "□",
        "A": "○",
        "B": "○",
        "X": "○",
        "Y": "○",
    }

    def __init__(self, term):

        self.term = term
        # Save a copy of the controls we can restore the
        # control text on deactivation
        self.DEFAULT_CONTROLS = self.CONTROLS.copy()

        self.CONTROL_RELEASE_TIMERS = self.CONTROLS.copy()
        for control in self.CONTROL_RELEASE_TIMERS.keys():
            self.CONTROL_RELEASE_TIMERS[control] = False

        self.auto_keypress_deactivation = True
        self.remote_connection = False

    def toggle_auto_keypress_deactivation(self, toggle):
        """Toggles whether or not the ControllerTUI should deactivate
        a control after a period of time.

        :param toggle: A True/False value that toggles auto keypress
        deactivation
        :type toggle: bool
        """

        self.auto_keypress_deactivation = toggle

    def set_remote_connection_status(self, status):
        """Sets whether or not the controller should render
        with remote connection specific controls.

        :param status: The status of the remote connection
        :type status: bool
        """

        self.remote_connection = status

    def activate_control(self, key, activated_text=None):

        if activated_text:
            self.CONTROLS[key] = activated_text
        else:
            self.CONTROLS[key] = self.term.bold_black_on_white(self.CONTROLS[key])

        # Keep track of when the key was pressed so we can release later
        if self.auto_keypress_deactivation:
            self.CONTROL_RELEASE_TIMERS[key] = time.perf_counter()

    def deactivate_control(self, key):

        self.CONTROLS[key] = self.DEFAULT_CONTROLS[key]

    def render_controller(self):

        if self.auto_keypress_deactivation:
            # Release any overdue timers
            for control in self.CONTROL_RELEASE_TIMERS.keys():
                pressed_time = self.CONTROL_RELEASE_TIMERS[control]
                current_time = time.perf_counter()
                if pressed_time is not False and current_time - pressed_time > 0.25:
                    self.deactivate_control(control)

        ZL = self.CONTROLS['ZL']
        L = self.CONTROLS['L']
        ZR = self.CONTROLS['ZR']
        R = self.CONTROLS['R']
        LU = self.CONTROLS['LS_UP']
        LL = self.CONTROLS['LS_LEFT']
        LR = self.CONTROLS['LS_RIGHT']
        LD = self.CONTROLS['LS_DOWN']
        RU = self.CONTROLS['RS_UP']
        RL = self.CONTROLS['RS_LEFT']
        RR = self.CONTROLS['RS_RIGHT']
        RD = self.CONTROLS['RS_DOWN']
        DU = self.CONTROLS['DPAD_UP']
        DL = self.CONTROLS['DPAD_LEFT']
        DR = self.CONTROLS['DPAD_RIGHT']
        DD = self.CONTROLS['DPAD_DOWN']
        MN = self.CONTROLS['MINUS']
        PL = self.CONTROLS['PLUS']
        HM = self.CONTROLS['HOME']
        CP = self.CONTROLS['CAPTURE']
        A = self.CONTROLS['A']
        B = self.CONTROLS['B']
        X = self.CONTROLS['X']
        Y = self.CONTROLS['Y']

        if self.remote_connection:
            lr_press = "L + R - - - - - - - - -▷ E"
        else:
            lr_press = "                          "

        print(self.term.home + self.term.move_y((self.term.height // 2) - 9))
        print(self.term.center(f"      {ZL}        {ZR}                                    "))
        print(self.term.center(f"    ─{L}──────────{R}─      ┌─────────────┬────────────┐"))
        print(self.term.center("  ╱                        ╲    │  Controls   │    Keys    │"))
        print(self.term.center(f" ╱   {LU}   {MN}       {PL}   {X}    ╲   └─────────────┴────────────┘"))  # noqa
        print(self.term.center(f"│   {LL}   {LR}    {CP}   {HM}   {Y}   {A}   │   Left Stick ─ ─ ─ ▷ W/A/S/D "))  # noqa
        print(self.term.center(f"│    {LD}               {B}     │   DPad ─ ─ ─ ─ ─ ─ ▷ G/V/B/N "))
        print(self.term.center(f"│        {DU}         {RU}       │   Capture/Home ─ ─ ─ ─ ▷ [/] "))
        print(self.term.center(f"│╲     {DL} □ {DR}      {RL}   {RR}     ╱│   +/- ─ ─ ─ ─ ─ ─ ─ ─ ─▷ 6/7 "))  # noqa
        print(self.term.center(f"│░░╲     {DD}         {RD}    ╱░░│   X/Y/B/A ─ ─ ─ ─ ─▷ J/I/K/L "))
        print(self.term.center("│░░░░╲ ──────────────── ╱░░░░│   L/ZL ─ ─ ─ ─ ─ ─ ─ ─ ▷ 1/2 "))
        print(self.term.center("│░░░░╱                  ╲░░░░│   R/ZR ─ ─ ─ ─ ─ ─ ─ ─ ▷ 8/9 "))
        print(self.term.center("│░░╱                      ╲░░│   Right Stick - - - ▷ Arrows "))
        print(self.term.center(f"│╱                          ╲│   {lr_press} "))


class InputTUI():

    KEYMAP = {
        # Left Stick Mapping
        "w": {
            "control": "LS_UP",
            "stick_data": {
                "stick_name": "L_STICK",
                "x": "+000",
                "y": "+100"
            }
        },
        "a": {
            "control": "LS_LEFT",
            "stick_data": {
                "stick_name": "L_STICK",
                "x": "-100",
                "y": "+000"
            }
        },
        "d": {
            "control": "LS_RIGHT",
            "stick_data": {
                "stick_name": "L_STICK",
                "x": "+100",
                "y": "+000"
            }
        },
        "s": {
            "control": "LS_DOWN",
            "stick_data": {
                "stick_name": "L_STICK",
                "x": "+000",
                "y": "-100"
            }
        },

        # Right Stick Mapping
        "KEY_UP": {
            "control": "RS_UP",
            "stick_data": {
                "stick_name": "R_STICK",
                "x": "+000",
                "y": "+100"
            }
        },
        "KEY_LEFT": {
            "control": "RS_LEFT",
            "stick_data": {
                "stick_name": "R_STICK",
                "x": "-100",
                "y": "+000"
            }
        },
        "KEY_RIGHT": {
            "control": "RS_RIGHT",
            "stick_data": {
                "stick_name": "R_STICK",
                "x": "+100",
                "y": "+000"
            }
        },
        "KEY_DOWN": {
            "control": "RS_DOWN",
            "stick_data": {
                "stick_name": "R_STICK",
                "x": "+000",
                "y": "-100"
            }
        },

        # Dpad Mapping
        "g": "DPAD_UP",
        "v": "DPAD_LEFT",
        "n": "DPAD_RIGHT",
        "b": "DPAD_DOWN",

        # Button Mapping
        "6": "MINUS",
        "7": "PLUS",
        "[": "CAPTURE",
        "]": "HOME",
        "i": "X",
        "j": "Y",
        "l": "A",
        "k": "B",

        # Triggers
        "1": "L",
        "2": "ZL",
        "8": "R",
        "9": "ZR",
    }

    def __init__(self, reconnect_target=None, debug=False, logfile=False, force_remote=False):

        self.reconnect_target = reconnect_target
        self.term = Terminal()
        if force_remote:
            self.remote_connection = True
        else:
            self.remote_connection = self.detect_remote_connection()
        self.controller = ControllerTUI(self.term)

        # Check if direct connection will fail
        if not self.remote_connection:
            try:
                from pynput import keyboard
            except ImportError as e:
                print("Unable to import pynput for direct input.")
                print("If you're accessing NXBT over a remote shell, ", end="")
                print("please use the 'remote_tui' option instead of 'tui'.")
                print("The original pynput import is displayed below:\n")
                print(e)
                exit(1)

        self.debug = debug
        self.logfile = logfile

    def detect_remote_connection(self):
        """Traverse up the parent processes and check if any
        have their parent as a remote daemon. If so, the python
        script is running under a remote connection.

        Remote shell detection is required for this TUI, due to
        keyboard input limitations on most remote connections.
        Specifically, no "keyup" events are sent when a key is
        released. Keyup events are required for proper input to
        the Switch, thus, we need to detect if the shell is a remote
        session and workaround this.

        :return: Returns a boolean value indicating whether or not
        the current script is running as SSH
        :rtype: bool
        """

        remote_connection = False
        remote_process_names = ['sshd', 'mosh-server']
        ppid = os.getppid()
        while ppid > 0:
            process = psutil.Process(ppid)
            if process.name() in remote_process_names:
                remote_connection = True
                break
            ppid = process.ppid()

        return remote_connection

    def start(self):

        self.mainloop(self.term)

    def mainloop(self, term):

        # Initializing a controller
        if not self.debug:
            self.nx = Nxbt(disable_logging=True)
        else:
            self.nx = Nxbt(debug=self.debug, logfile=self.logfile)
        self.controller_index = self.nx.create_controller(
            PRO_CONTROLLER,
            reconnect_address=self.reconnect_target)

        state = None
        spinner = LoadingSpinner()
        errors = None
        try:
            with term.cbreak(), term.keypad(), term.location(), term.hidden_cursor():
                print(term.home + term.clear)
                self.render_top_bar(term)
                self.render_bottom_bar(term)
                self.render_start_screen(term, "Loading")
                inp = term.inkey(timeout=0)

                # Loading Screen
                while inp != chr(113):  # Checking for q press
                    # Check key at 15hz
                    inp = term.inkey(timeout=1/30)
                    new_state = self.nx.state[self.controller_index]["state"]

                    if new_state != state:
                        state = new_state

                        loading_text = "Loading"
                        if state == "initializing":
                            loading_text = "Initializing Controller"
                        elif state == "connecting":
                            loading_text = "Connecting to any Nintendo Switch"
                        elif state == "reconnecting":
                            loading_text = "Reconnecting to Nintendo Switch"
                        elif state == "connected":
                            loading_text = "Connected!"
                        elif state == "crashed":
                            errors = self.nx.state[self.controller_index]["errors"]
                            exit(1)
                        self.render_start_screen(term, loading_text)

                    print(term.move_y((term.height // 2) + 6))
                    if state != "connected":
                        print(term.bold(term.center(spinner.get_spinner_char())))
                    else:
                        print(term.center(""))

                    if state == "connected":
                        time.sleep(1)
                        break

                # Main Gamepad Input Loop
                if state == "connected":
                    if self.remote_connection:
                        self.remote_input_loop(term)
                    else:
                        self.direct_input_loop(term)

        except KeyboardInterrupt:
            pass
        finally:
            print(term.clear())
            if errors:
                print("The TUI encountered the following errors:")
                print(errors)

    def remote_input_loop(self, term):

        self.controller.set_remote_connection_status(True)

        inp = term.inkey(timeout=0)
        while inp != chr(113):  # Checking for q press
            # Cutoff large buffered input from the deque
            # so that we avoid spamming the Switch after
            # a key releases from being held.
            # Increasing the size of the buffer does not
            # smooth out the jagginess of input.
            if len(term._keyboard_buf) > 1:
                term._keyboard_buf = deque([term._keyboard_buf.pop()])

            inp = term.inkey(1/66)

            pressed_key = None
            if inp.is_sequence:
                pressed_key = inp.name
            elif inp:
                pressed_key = inp

            if pressed_key == 'e':
                self.controller.activate_control('L')
                self.controller.activate_control('R')
                self.nx.macro(self.controller_index, "L R 0.1s")
            else:
                try:
                    control_data = self.KEYMAP[pressed_key]
                    if type(control_data) == dict and "stick_data" in control_data.keys():
                        x_value = control_data['stick_data']['x']
                        y_value = control_data['stick_data']['y']
                        stick_name = control_data['stick_data']['stick_name']

                        self.controller.activate_control(control_data["control"])
                        self.nx.macro(
                            self.controller_index,
                            f"{stick_name}@{x_value}{y_value} 0.1s")
                    else:
                        self.controller.activate_control(control_data)
                        self.nx.macro(self.controller_index, f"{control_data} 0.05s")
                except KeyError:
                    pass

            self.controller.render_controller()

            self.check_for_disconnect(term)

    def direct_input_loop(self, term):
        from pynput import keyboard

        self.controller.toggle_auto_keypress_deactivation(False)
        self.exit_tui = False
        self.capture_input = True

        # Create a packet that is accessible from a multiprocessing Process
        # and from within threads
        packet_manager = multiprocessing.Manager()
        input_packet = packet_manager.dict()
        input_packet["packet"] = self.nx.create_input_packet()

        print(term.move_y(term.height - 5))
        print(term.center(term.bold_black_on_white(" <Press esc to toggle input capture> ")))

        def on_press(key):

            # Parse the key press event
            pressed_key = None
            try:
                pressed_key = key.char
            except AttributeError:
                pressed_key = str(key).replace(".", "_").upper()

            if not self.capture_input:  # If we're not capturing input, pass
                pass
            else:
                try:
                    control_data = self.KEYMAP[pressed_key]
                    packet = input_packet["packet"]
                    if type(control_data) == dict and "stick_data" in control_data.keys():
                        stick_name = control_data['stick_data']['stick_name']
                        self.controller.activate_control(control_data["control"])
                        packet[stick_name][control_data["control"]] = True
                    else:
                        self.controller.activate_control(control_data)
                        packet[control_data] = True
                    input_packet["packet"] = packet
                except KeyError:
                    pass

        def on_release(key):

            # Parse the key release event
            released_key = None
            try:
                released_key = key.char
            except AttributeError:
                released_key = str(key).replace(".", "_").upper()

            # If the esc key is released, toggle input capturing
            if released_key == "KEY_ESC":
                self.capture_input = not self.capture_input

            # Exit on q key press
            if released_key == 'q':
                self.exit_tui = True
                return False

            if not self.capture_input:  # If we're not capturing input, pass
                pass
            else:
                try:
                    control_data = self.KEYMAP[released_key]
                    packet = input_packet["packet"]
                    if type(control_data) == dict and "stick_data" in control_data.keys():
                        stick_name = control_data['stick_data']['stick_name']
                        self.controller.deactivate_control(control_data["control"])
                        packet[stick_name][control_data["control"]] = False
                    else:
                        self.controller.deactivate_control(control_data)
                        packet[control_data] = False
                    input_packet["packet"] = packet
                except KeyError:
                    pass

        def input_worker(nxbt, controller_index, input_packet):

            while True:
                packet = input_packet["packet"]

                # Calculating left x/y stick values
                ls_x_value = 0
                ls_y_value = 0
                if packet["L_STICK"]["LS_LEFT"]:
                    ls_x_value -= 100
                if packet["L_STICK"]["LS_RIGHT"]:
                    ls_x_value += 100
                if packet["L_STICK"]["LS_UP"]:
                    ls_y_value += 100
                if packet["L_STICK"]["LS_DOWN"]:
                    ls_y_value -= 100
                packet["L_STICK"]["X_VALUE"] = ls_x_value
                packet["L_STICK"]["Y_VALUE"] = ls_y_value

                # Calculating right x/y stick values
                rs_x_value = 0
                rs_y_value = 0
                if packet["R_STICK"]["RS_LEFT"]:
                    rs_x_value -= 100
                if packet["R_STICK"]["RS_RIGHT"]:
                    rs_x_value += 100
                if packet["R_STICK"]["RS_UP"]:
                    rs_y_value += 100
                if packet["R_STICK"]["RS_DOWN"]:
                    rs_y_value -= 100
                packet["R_STICK"]["X_VALUE"] = rs_x_value
                packet["R_STICK"]["Y_VALUE"] = rs_y_value

                nxbt.set_controller_input(controller_index, packet)
                time.sleep(1/120)

        input_process = multiprocessing.Process(
            target=input_worker, args=(self.nx, self.controller_index, input_packet))
        input_process.start()

        # Start a non-blocking keyboard event listener
        listener = keyboard.Listener(
            on_press=on_press,
            on_release=on_release)
        listener.start()

        # Main TUI Loop
        while True:
            if self.exit_tui:
                packet_manager.shutdown()
                input_process.terminate()
                break
            if not self.capture_input:
                print(term.home + term.move_y((term.height // 2) - 4))
                print(term.bold_black_on_white(term.center("")))
                print(term.bold_black_on_white(term.center(
                    "<Input Paused. Press ESC Again to Begin Capturing Input>"
                )))
                print(term.bold_black_on_white(term.center("")))
            else:
                self.controller.render_controller()
            self.check_for_disconnect(term)
            time.sleep(1/120)

    def render_start_screen(self, term, loading_text):

        print(term.home + term.move_y((term.height // 2) - 8))
        print(term.center("___╲╱___"))
        print(term.center("│╲  ╱╲  ╱│"))
        print(term.center("│ ╲╱__╲╱ │"))
        print(term.center("│╱      ╲│"))
        print(term.center("┌──────────────────┐"))
        print(term.center("│     NXBT TUI     │"))
        print(term.center("└──────────────────┘"))
        print(term.center(""))
        print(term.black_on_white(term.center("")))
        print(term.bold_black_on_white(term.center(loading_text)))
        print(term.black_on_white(term.center("")))

    def render_top_bar(self, term):

        print(term.move_y(1))
        if self.remote_connection:
            print(term.bold_black_on_white(term.center(term.bold_black_on_red("  REMOTE MODE  "))))
            warning = " WARNING: MACROS WILL BE USED ON KEYPRESS DUE TO REMOTE CLI LIMITATIONS "
            print(term.center(term.black_on_red(warning)))
        else:
            print(term.bold_black_on_white(term.center("DIRECT INPUT MODE")))
        print(term.move_y(1))
        print(term.white_on_black(" NXBT TUI 🎮 "))

    def render_bottom_bar(self, term):

        print(term.move_y(term.height))
        print(term.center(term.bold_black_on_white(" <Press q to quit> ")))

    def check_for_disconnect(self, term):

        state = self.nx.state[self.controller_index]["state"]
        if state != 'connected':
            print(term.home + term.move_y((term.height // 2) - 4))
            print(term.bold_black_on_red(term.center("")))
            print(term.bold_black_on_red(term.center(state.title())))
            print(term.bold_black_on_red(term.center("")))

            if state == 'crashed':
                time.sleep(3)
                term.clear()
                errors = self.nx.state[self.controller_index]["errors"]
                raise ConnectionError(errors)

            while True:
                inp = term.inkey(1/30)
                if inp == chr(113):
                    exit(1)
                elif self.nx.state[self.controller_index]["state"] == 'connected':
                    break


def main():
    """Program entry point."""
    tui = InputTUI()
    tui.start()


if __name__ == '__main__':
    main()



import asyncio

class _Request:
    def __init__(self, value, loop):
        self.value = value
        self.future = loop.create_future()


class MySemaphore(asyncio.Semaphore):
    """
    An implementation of the asyncio-Semaphore with a few more features.
    Most this code is copied from the original CPython implementation.
    """
    def __init__(self, value):
        super().__init__(value)
        self._value = value
        self._waiters = [] # Normal people would use an actual Queue. The standard queues are shit
        self._aquired = 0

    def _check_next(self):
        while self._waiters and (self._waiters[0].future.done() or self._value >= self._waiters[0].value):
            r = self._waiters.pop(0)
            if not r.future.done():
                r.future.set_result(None)
                return

    async def acquire(self, count=1):
        if count < 0:
            raise ValueError("Semaphore acquire with count < 0")
        while self._value < count:
            r = _Request(count, self._loop)
            self._waiters.append(r)
            try:
                await r.future
            except:
                r.future.cancel()
                # original has an if here, we wont take the call anymore, call the next one
                self._check_next()
                raise
        self._aquired += count
        self._value -= count
        self._check_next()
        return True

    def reduce(self, value):
        self._value -= value

    def increase(self, value):
        self._value += value
        self._check_next()

    def get_value(self):
        return self._value

    def get_aquired(self):
        return self._aquired

    def release(self, count=1):
        if count < 0:
            raise ValueError("Semaphore release with 0 < count")
        self._value += count
        self._aquired -= count
        self._check_next()

class MyBoundedSemaphore(MySemaphore):
    """
    Äquivalent to asyncio.BoundedSemaphore,
    also with more features
    """
    def __init__(self, limit=1, value=None):
        super().__init__(value if not value is None else limit)
        self._limit = limit

    def get_limit(self):
        return self._limit

    def set_limit(self, value):
        self._limit = value
        self._value = min(self._value, self._limit)

    def release(self, count=1, best_effort=False):
        if self._value + count > self._limit:
            if best_effort:
                count = self._limit - self._value
            else:
                raise ValueError('BoundedSemaphore released too many times')
        super().release(count)




class FlashMemory:
    def __init__(self, spi_flash_memory_data=None, default_stick_cal=False, size=0x80000):
        """
        :param spi_flash_memory_data: data from a memory dump (can be created using dump_spi_flash.py).
        :param default_stick_cal: If True, override stick calibration bytes with factory default
        :param size of the memory dump, should be constant
        """
        if spi_flash_memory_data is None:
            spi_flash_memory_data = [0xFF] * size  # Blank data is all 0xFF
            default_stick_cal = True

        if len(spi_flash_memory_data) != size:
            raise ValueError(f'Given data size {len(spi_flash_memory_data)} does not match size {size}.')
        if isinstance(spi_flash_memory_data, bytes):
            spi_flash_memory_data = list(spi_flash_memory_data)

        # set default controller stick calibration
        if default_stick_cal:
            # L-stick factory calibration
            spi_flash_memory_data[0x603D:0x6046] = [0x00, 0x07, 0x70, 0x00, 0x08, 0x80, 0x00, 0x07, 0x70]
            # R-stick factory calibration
            spi_flash_memory_data[0x6046:0x604F] = [0x00, 0x08, 0x80, 0x00, 0x07, 0x70, 0x00, 0x07, 0x70]

        self.data = spi_flash_memory_data

    def __getitem__(self, item):
        return self.data[item]

    def get_factory_l_stick_calibration(self):
        """
        :returns 9 left stick factory calibration bytes
        """
        return self.data[0x603D:0x6046]

    def get_factory_r_stick_calibration(self):
        """
        :returns 9 right stick factory calibration bytes
        """
        return self.data[0x6046:0x604F]

    def get_user_l_stick_calibration(self):
        """
        :returns 9 left stick user calibration bytes if the data is available, otherwise None
        """
        # check if calibration data is available:
        if self.data[0x8010] == 0xB2 and self.data[0x8011] == 0xA1:
            return self.data[0x8012:0x801B]
        else:
            return None

    def get_user_r_stick_calibration(self):
        """
        :returns 9 right stick user calibration bytes if the data is available, otherwise None
        """
        # check if calibration data is available:
        if self.data[0x801B] == 0xB2 and self.data[0x801C] == 0xA1:
            return self.data[0x801D:0x8026]
        else:
            return None



from multiprocessing import Process, Lock, Queue, Manager
import queue
from enum import Enum
import atexit
import signal
import os
import sys
import time
import json

import dbus

from .controller import ControllerServer
from .controller import ControllerTypes
from .bluez import BlueZ, find_objects, toggle_clean_bluez
from .bluez import replace_mac_addresses
from .bluez import find_devices_by_alias
from .bluez import SERVICE_NAME, ADAPTER_INTERFACE
from .logging import create_logger


JOYCON_L = ControllerTypes.JOYCON_L
JOYCON_R = ControllerTypes.JOYCON_R
PRO_CONTROLLER = ControllerTypes.PRO_CONTROLLER


DIRECT_INPUT_PACKET = {
    # Sticks
    "L_STICK": {
        "PRESSED": False,
        "X_VALUE": 0,
        "Y_VALUE": 0,
        # Keyboard position calculation values
        "LS_UP": False,
        "LS_LEFT": False,
        "LS_RIGHT": False,
        "LS_DOWN": False
    },
    "R_STICK": {
        "PRESSED": False,
        "X_VALUE": 0,
        "Y_VALUE": 0,
        # Keyboard position calculation values
        "RS_UP": False,
        "RS_LEFT": False,
        "RS_RIGHT": False,
        "RS_DOWN": False
    },
    # Dpad
    "DPAD_UP": False,
    "DPAD_LEFT": False,
    "DPAD_RIGHT": False,
    "DPAD_DOWN": False,
    # Triggers
    "L": False,
    "ZL": False,
    "R": False,
    "ZR": False,
    # Joy-Con Specific Buttons
    "JCL_SR": False,
    "JCL_SL": False,
    "JCR_SR": False,
    "JCR_SL": False,
    # Meta buttons
    "PLUS": False,
    "MINUS": False,
    "HOME": False,
    "CAPTURE": False,
    # Buttons
    "Y": False,
    "X": False,
    "B": False,
    "A": False
}


class Buttons():
    """The button object containing the button string constants.
    """

    Y = 'Y'
    X = 'X'
    B = 'B'
    A = 'A'
    JCL_SR = 'JCL_SR'
    JCL_SL = 'JCL_SL'
    R = 'R'
    ZR = 'ZR'
    MINUS = 'MINUS'
    PLUS = 'PLUS'
    R_STICK_PRESS = 'R_STICK_PRESS'
    L_STICK_PRESS = 'L_STICK_PRESS'
    HOME = 'HOME'
    CAPTURE = 'CAPTURE'
    DPAD_DOWN = 'DPAD_DOWN'
    DPAD_UP = 'DPAD_UP'
    DPAD_RIGHT = 'DPAD_RIGHT'
    DPAD_LEFT = 'DPAD_LEFT'
    JCR_SR = 'JCR_SR'
    JCR_SL = 'JCR_SL'
    L = 'L'
    ZL = 'ZL'


class Sticks():
    """The sticks object containing the joystick string constants.
    """

    RIGHT_STICK = "R_STICK"
    LEFT_STICK = "L_STICK"


class NxbtCommands(Enum):
    """An enumeration containing the nxbt message
    commands.
    """

    CREATE_CONTROLLER = 0
    INPUT_MACRO = 1
    STOP_MACRO = 2
    CLEAR_MACROS = 3
    CLEAR_ALL_MACROS = 4
    REMOVE_CONTROLLER = 5
    QUIT = 6


class Nxbt():
    """The nxbt object implements the core multiprocessing logic
    and message passing API that acts as the central of the application.
    Upon creation, a multiprocessing Process is spun off to act at the
    manager for all emulated Nintendo Switch controllers. Messages
    are passed into a queue which is consumed and acted upon by the
    _command_manager.

    All function calls that interact or control the emulated controllers
    are simply message constructors that submit to the central task_queue.
    This allows for thread-safe control of emulated controllers.
    """

    def __init__(self, debug=False, log_to_file=False, disable_logging=False):
        """Initializes the necessary multiprocessing resources and starts
        the multiprocessing processes.

        :param debug: Enables the debugging functionality of
        nxbt, defaults to False
        :type debug: bool, optional
        :param log_to_file: A boolean value that indiciates whether or not
        a log should be saved to the current working directory, defaults to False
        :type log_to_file: bool, optional
        :param disable_logging: Routes all logging calls to a null log handler.
        :type disable_logging: bool, optional, defaults to False.
        """

        self.debug = debug
        self.logger = create_logger(
            debug=self.debug, log_to_file=log_to_file, disable_logging=disable_logging)

        # Main queue for nbxt tasks
        self.task_queue = Queue()

        # Sychronizes bluetooth actions
        self._bluetooth_lock = Lock()

        # Creates/manages shared resources
        self.resource_manager = Manager()
        # Shared dictionary for viewing overall nxbt state.
        # Should treated as read-only except by
        # the main nxbt multiprocessing process.
        self.manager_state = self.resource_manager.dict()
        self.manager_state_lock = Lock()

        # Shared, controller management properties.
        # The controller lock is used to sychronize use.
        self._controller_lock = Lock()
        self._controller_counter = 0
        self._adapters_in_use = {}
        self._controller_adapter_lookup = {}

        # Disable the BlueZ input plugin so we can use the
        # HID control/interrupt Bluetooth ports
        toggle_clean_bluez(True)

        # Exit handler
        atexit.register(self._on_exit)

        # Starting the nxbt worker process
        self.controllers = Process(
            target=self._command_manager,
            args=((self.task_queue), (self.manager_state)))
        # Disabling daemonization since we need to spawn
        # other controller processes, however, this means
        # we need to cleanup on exit.
        self.controllers.daemon = False
        self.controllers.start()

    def _on_exit(self):
        """The exit handler function used with the atexit module.
        This function attempts to gracefully exit by terminating
        all spun up multiprocessing Processes. This is done to
        ensure no zombie processes linger after exit.
        """

        # Need to explicitly kill the controllers process
        # since it isn't daemonized.
        if hasattr(self, "controllers") and self.controllers.is_alive():
            self.controllers.terminate()

        self.resource_manager.shutdown()

        # Re-enable the BlueZ plugins, if we have permission
        toggle_clean_bluez(False)

    def _command_manager(self, task_queue, state):
        """Used as the main multiprocessing Process that is launched
        on startup to handle the message passing and instantiation of
        the controllers. Messages are pulled out of a Queue and passed
        as appropriately phrased function calls to the ControllerManager.

        :param task_queue: A multiprocessing Queue used as the source
        of messages
        :type task_queue: multiprocessing.Queue
        :param state: A dict used to store the shared state of the
        emulated controllers.
        :type state: multiprocessing.Manager().dict
        """

        cm = _ControllerManager(state, self._bluetooth_lock)
        # Ensure a SystemExit exception is raised on SIGTERM
        # so that we can gracefully shutdown.
        signal.signal(signal.SIGTERM, lambda sigterm_handler: sys.exit(0))

        try:
            while True:
                try:
                    msg = task_queue.get(timeout=5)
                except queue.Empty:
                    msg = None

                if msg:
                    if msg["command"] == NxbtCommands.CREATE_CONTROLLER:
                        cm.create_controller(
                            msg["arguments"]["controller_index"],
                            msg["arguments"]["controller_type"],
                            msg["arguments"]["adapter_path"],
                            msg["arguments"]["colour_body"],
                            msg["arguments"]["colour_buttons"],
                            msg["arguments"]["reconnect_address"])
                    elif msg["command"] == NxbtCommands.INPUT_MACRO:
                        cm.input_macro(
                            msg["arguments"]["controller_index"],
                            msg["arguments"]["macro"],
                            msg["arguments"]["macro_id"])
                    elif msg["command"] == NxbtCommands.STOP_MACRO:
                        cm.stop_macro(
                            msg["arguments"]["controller_index"],
                            msg["arguments"]["macro_id"])
                    elif msg["command"] == NxbtCommands.CLEAR_MACROS:
                        cm.clear_macros(
                            msg["arguments"]["controller_index"])
                    elif msg["command"] == NxbtCommands.REMOVE_CONTROLLER:
                        index = msg["arguments"]["controller_index"]
                        cm.clear_macros(index)
                        cm.remove_controller(index)

        finally:
            cm.shutdown()
            sys.exit(0)

    def macro(self, controller_index, macro, block=True):
        """Used to input a given macro on a specified controller.
        This is done by creating and passing an INPUT_MACRO
        message into the task queue with the given macro.

        If block is set to True, this function waits until the
        macro_id (generated on the submission of the macro)
        shows up under the "finished_macros" list communicated
        under the controllers shared state.

        :param controller_index: The index of a given controller
        :type controller_index: int
        :param macro: The series of button presses and timings
        to be passed to the controller
        :type macro: string
        :param block: A boolean variable indicating whether or not
        to block until the macro completes, defaults to True
        :type block: bool, optional
        :raises ValueError: If the controller_index does not exist
        :return: The generated ID of the passed macro. This ID
        will show up under the "finished_macros" list communicated
        in the controllers shared state.
        :rtype: str
        """

        if controller_index not in self.manager_state.keys():
            raise ValueError("Specified controller does not exist")

        # Get a unique ID to identify the macro
        # so we can check when the controller is done inputting it
        macro_id = os.urandom(24).hex()
        self.task_queue.put({
            "command": NxbtCommands.INPUT_MACRO,
            "arguments": {
                "controller_index": controller_index,
                "macro": macro,
                "macro_id": macro_id,
            }
        })

        if block:
            while True:
                finished = (self.manager_state
                            [controller_index]["finished_macros"])
                if macro_id in finished:
                    break

                time.sleep(1/120)  # Wait one Pro Controller cycle

        return macro_id

    def press_buttons(self, controller_index, buttons, down=0.1, up=0.1, block=True):
        """Used to press a given set of buttons on the controller for a
        specified up and down duration. This is done by inputting a macro
        configured with the specified button presses and timings.

        :param controller_index: The index of a given controller
        :type controller_index: int
        :param buttons: A list of nxbt.Buttons
        :type buttons: list
        :param down: How long to hold the buttons down for
        in seconds, defaults to 0.1
        :type down: float, optional
        :param up: How long to release the button for
        in seconds, defaults to 0.1
        :type up: float, optional
        :param block: A boolean variable indicating whether or not
        to block until the macro completes, defaults to True
        :type block: bool, optional
        :return: The generated ID of the passed macro. This ID
        will show up under the "finished_macros" list communicated
        in the controllers shared state.
        :rtype: str
        """

        macro_buttons = " ".join(buttons)
        macro_times = f"{down}s \n{up}s"
        macro = macro_buttons + " " + macro_times

        macro_id = self.macro(controller_index, macro, block=block)

        return macro_id

    def tilt_stick(self, controller_index, stick, x, y,
                   tilted=0.1, released=0.1, block=True):
        """Used to tilt a given stick on the controller for a
        specified tilted and released duration. This is done by
        inputting a macro configured with the specified stick tilts
        and timings.

        :param controller_index: The index of a given controller
        :type controller_index: int
        :param stick: The right or left nxbt.Stick
        :type stick: nxbt.Stick
        :param x: The positive or negative X-Axis of the stick on
        a 0 to 100 scale
        :type x: int
        :param y: The positive or negative Y-Axis of the stick on
        a 0 to 100 scale
        :type y: int
        :param tilted: The time the stick should remain tilted
        for, defaults to 0.1
        :type tilted: float, optional
        :param released: The time the stick should remain
        released for, defaults to 0.1
        :type released: float, optional
        :type block: bool, optional
        :return: The generated ID of the passed macro. This ID
        will show up under the "finished_macros" list communicated
        in the controllers shared state.
        :rtype: str
        """

        if controller_index not in self.manager_state.keys():
            raise ValueError("Specified controller does not exist")

        if x >= 0:
            x_parsed = f'+{x:03}'
        else:
            x_parsed = f'{x:04}'

        if y >= 0:
            y_parsed = f'+{y:03}'
        else:
            y_parsed = f'{y:04}'

        macro = f'{stick}@{x_parsed}{y_parsed} {tilted}s\n{released}s'

        macro_id = self.macro(controller_index, macro, block=block)

        return macro_id

    def stop_macro(self, controller_index, macro_id, block=True):
        """Used to stop a given macro by its macro ID. After
        the macro has been stopped, its macro ID will show up
        as a finished macro in the respective controllers
        "finished_macros" list communicated in its state.

        :param controller_index: The index of a given controller
        :type controller_index: int
        :param macro_id: The ID of a given macro (queued or running)
        :type macro_id: str
        :param block: A boolean variable indicating whether or not
        to block until the macro is stopped, defaults to True
        :type block: bool, optional
        :raises ValueError: If the controller_index does not exist
        """

        if controller_index not in self.manager_state.keys():
            raise ValueError("Specified controller does not exist")

        self.task_queue.put({
            "command": NxbtCommands.STOP_MACRO,
            "arguments": {
                "controller_index": controller_index,
                "macro_id": macro_id,
            }
        })

        if block:
            while True:
                finished = (self.manager_state
                            [controller_index]["finished_macros"])
                if macro_id in finished:
                    break

                time.sleep(1/120)

    def clear_macros(self, controller_index):
        """Clears all running and queued macros on a specified
        controller.

        WARNING: Any blocking macro calls will continue to run
        forever if this command is run.

        :param controller_index: The index of a given controller
        :type controller_index: int
        :raises ValueError: If the controller_index does not exist
        """

        if controller_index not in self.manager_state.keys():
            raise ValueError("Specified controller does not exist")

        self.task_queue.put({
            "command": NxbtCommands.CLEAR_MACROS,
            "arguments": {
                "controller_index": controller_index,
            }
        })

    def clear_all_macros(self):
        """Clears all running and queued macros on all
        controllers.
        """

        for controller in self.manager_state.keys():
            self.clear_macros(controller)

    def set_controller_input(self, controller_index, input_packet):
        """Sets the controllers buttons and analog sticks for 1 cycle.
        This means that exactly 1 packet will be sent to the Switch with
        input specified with this method. To keep a continuous input
        stream of a desired input, packets must be set at a rate that
        roughly matches the set controller. Eg: An emulated Pro Controller's
        input must be set at roughly 120Hz and a Joy-Con's at 60Hz.

        :param controller_index: The index of the emulated controller
        :type controller_index: int
        :param input_packet: The input packet with the desired input. This
        *must* be an instance of the create_input_packet method.
        :type input_packet: dict
        :raises ValueError: On bad controller index
        """

        if controller_index not in self.manager_state.keys():
            raise ValueError("Specified controller does not exist")

        self.manager_state[controller_index]["direct_input"] = input_packet

    def create_input_packet(self):
        """Creates an input packet that is used to specify the input
        of a controller for a single cycle.

        :return: An input packet dictionary
        :rtype: dict
        """

        # Create a copy of the direct input packet in a thread safe manner.
        # NOTE: Using the copy.deepcopy method of copying dicts IS NOT thread safe.
        return json.loads(json.dumps(DIRECT_INPUT_PACKET))

    def create_controller(self, controller_type, adapter_path=None,
                          colour_body=None, colour_buttons=None,
                          reconnect_address=None):
        """Used to create a Nintendo Switch controller of a
        given type and colour on an (optionally) specified
        bluetooth adapter.

        If no Bluetooth adapter is specified, the first available
        adapter is used.

        If the reconnect_address is specified, the controller
        will attempt to reconnect to the Switch, rather than
        simply letting any Switch connect to it. To ensure
        that the reconnect succeeds, the Switch must be on
        and *not* on the Change Grip/Order menu.

        :param controller_type: The type of controller to create
        :type controller_type: ControllerTypes
        :param adapter_path: The DBus path to a given Bluetooth
        adapter, defaults to None
        :type adapter_path: str, optional
        :param colour_body: The body colour of the controller
        represented by a hexadecimal colour value (a list of
        three ints (0-255)), defaults to None
        :type colour_body: list, optional
        :param colour_buttons: The button colour of the controller
        represented by a hexadecimal colour value (a list of
        three ints (0-255)), defaults to None
        :type colour_buttons: list, optional
        :param reconnect_address: A previously connected to
        Switch's Bluetooth MAC address, defaults to None
        :type reconnect_address: str or list, optional
        :raises ValueError: If specified adapter is unavailable
        :raises ValueError: If specified adapter is in use
        :return: The index of the created controller
        :rtype: int
        """
        if adapter_path:
            if adapter_path not in self.get_available_adapters():
                raise ValueError("Specified adapter is unavailable")

            if adapter_path in self._adapters_in_use.keys():
                raise ValueError("Specified adapter in use")
        else:
            # Get all adapters we can use
            usable_adapters = list(
                set(self.get_available_adapters()) - set(self._adapters_in_use))
            if len(usable_adapters) > 0:
                # Use the first available adapter
                adapter_path = usable_adapters[0]
            else:
                raise ValueError("No adapters available")

        controller_index = None
        try:
            self._controller_lock.acquire()
            self.task_queue.put({
                "command": NxbtCommands.CREATE_CONTROLLER,
                "arguments": {
                    "controller_index": self._controller_counter,
                    "controller_type": controller_type,
                    "adapter_path": adapter_path,
                    "colour_body": colour_body,
                    "colour_buttons": colour_buttons,
                    "reconnect_address": reconnect_address,
                }
            })
            controller_index = self._controller_counter
            self._controller_counter += 1
            self._adapters_in_use[adapter_path] = controller_index
            self._controller_adapter_lookup[controller_index] = adapter_path

            # Block until the controller is ready
            # This needs to be done to prevent race conditions
            # on Bluetooth resources.
            if type(controller_index) == int:
                while True:
                    if controller_index in self.manager_state.keys():
                        state = self.manager_state[controller_index]
                        if (state["state"] == "connecting" or
                                state["state"] == "reconnecting" or
                                state["state"] == "crashed"):
                            break

                    time.sleep(1/30)
        finally:
            self._controller_lock.release()

        return controller_index

    def remove_controller(self, controller_index):
        """Terminates and removes a given controller.

        :param controller_index: The index of a given controller
        :type controller_index: int
        :raises ValueError: If controller does not exist
        """

        if controller_index not in self.manager_state.keys():
            if controller_index in self._controller_adapter_lookup.keys():
                # Attempt to free any adapters claimed by a crashed controller
                try:
                    adapter_path = self._controller_adapter_lookup.pop(controller_index, None)
                    self._adapters_in_use.pop(adapter_path, None)
                except Exception:
                    pass
            raise ValueError("Specified controller does not exist")

        self._controller_lock.acquire()
        try:
            adapter_path = self._controller_adapter_lookup.pop(controller_index, None)
            self._adapters_in_use.pop(adapter_path, None)
        finally:
            self._controller_lock.release()

        self.task_queue.put({
            "command": NxbtCommands.REMOVE_CONTROLLER,
            "arguments": {
                "controller_index": controller_index,
            }
        })

    def wait_for_connection(self, controller_index):
        """Blocks until a given controller is connected
        to a Nintendo Switch.

        :param controller_index: The index of a given controller
        :type controller_index: int
        """

        while not self.state[controller_index]["state"] == "connected":
            if self.state[controller_index]["state"] == "crashed":
                raise OSError("The watched controller has crashe with error",
                              self.state[controller_index]["errors"])
            pass

    def get_available_adapters(self):
        """Gets the DBus paths of all available Bluetooth
        adapters.

        :return: A list of available adapter paths
        :rtype: list
        """

        bus = dbus.SystemBus()
        adapters = find_objects(bus, SERVICE_NAME, ADAPTER_INTERFACE)
        bus.close()

        return adapters

    def get_switch_addresses(self):
        """Gets the Bluetooth MAC addresses of all
        previously connected Nintendo Switchs

        :return: A list of Bluetooth MAC addresses
        :rtype: list
        """

        return (find_devices_by_alias("Nintendo Switch"))

    @property
    def state(self):
        """The state of all created and running controllers.
        This state is read-only and is represented as a dict.
        The state dict's structure follows:

        {
            "controller_index"
                {
                    "state":
                        "initializing" or
                        "connecting" or
                        "reconnecting" or
                        "connected" or
                        "crashed"
                    "finished_macros":
                        A list of UUIDs
                    "errors":
                        A string with the crash error
                    "direct_input":
                        A dictionary that represents all inputs
                        being directly input into the controller.
                }
        }

        :return: The state dict
        :rtype: dict
        """

        return self.manager_state


class _ControllerManager():
    """Used as the manager for all controllers. Each controller is
    a daemon multiprocessing Process that the ControllerManager
    object creates and manages.

    The ControllerManager object submits messages to the respective
    queues of each controller process for tasks such as macro submission
    or macro clearing/stopping.
    """

    def __init__(self, state, lock):

        self.state = state
        self.lock = lock
        self.controller_resources = Manager()
        self._controller_queues = {}
        self._children = {}

    def create_controller(self, index, controller_type, adapter_path,
                          colour_body=None, colour_buttons=None,
                          reconnect_address=None):
        """Instantiates a given controller as a multiprocessing
        Process with a shared state dict and a task queue.

        Configuration options are available in the form of
        controller colours.

        :param index: The index of the controller
        :type index: int
        :param controller_type: The type of Nintendo Switch controller
        :type controller_type: ControllerTypes
        :param adapter_path: The DBus path to the Bluetooth adapter
        :type adapter_path: str
        :param colour_body: A list of three ints representing the hex
        colour of the controller, defaults to None
        :type colour_body: list, optional
        :param colour_buttons: A list of three ints representing the
        hex colour of the controller, defaults to None
        :type colour_buttons: list, optional
        :param reconnect_address: The address of a Nintendo Switch
        to reconnect to, defaults to None
        :type reconnect_address: str, optional
        """

        controller_queue = Queue()

        controller_state = self.controller_resources.dict()
        controller_state["state"] = "initializing"
        controller_state["finished_macros"] = []
        controller_state["errors"] = False
        controller_state["direct_input"] = json.loads(json.dumps(DIRECT_INPUT_PACKET))
        controller_state["colour_body"] = colour_body
        controller_state["colour_buttons"] = colour_buttons
        controller_state["type"] = str(controller_type)
        controller_state["adapter_path"] = adapter_path
        controller_state["last_connection"] = None

        self._controller_queues[index] = controller_queue

        self.state[index] = controller_state

        server = ControllerServer(controller_type,
                                  adapter_path=adapter_path,
                                  lock=self.lock,
                                  state=controller_state,
                                  task_queue=controller_queue,
                                  colour_body=colour_body,
                                  colour_buttons=colour_buttons)
        controller = Process(target=server.run, args=(reconnect_address,))
        controller.daemon = True
        self._children[index] = controller
        controller.start()

    def input_macro(self, index, macro, macro_id):

        self._controller_queues[index].put({
            "type": "macro",
            "macro": macro,
            "macro_id": macro_id
        })

    def stop_macro(self, index, macro_id):

        self._controller_queues[index].put({
            "type": "stop",
            "macro_id": macro_id,
        })

    def clear_macros(self, index):

        self._controller_queues[index].put({
            "type": "clear",
        })

    def remove_controller(self, index):
        
        self._children[index].terminate()
        self.state.pop(index, None)

    def shutdown(self):

        # Loop over children and kill all
        for index in self._children.keys():
            child = self._children[index]
            child.terminate()

        self.controller_resources.shutdown()



from abc import abstractmethod
from enum import Enum
from typing import Union


class Button(Enum):
    A = 0
    B = 1
    X = 2
    Y = 3
    DPAD_UP = 4
    DPAD_DOWN = 5
    DPAD_LEFT = 6
    DPAD_RIGHT = 7
    L_STICK_PRESS = 8
    R_STICK_PRESS = 9
    SHOULDER_L = 10
    SHOULDER_R = 11
    SHOULDER_ZL = 12
    SHOULDER_ZR = 13
    HOME = 14
    CAPTURE = 15
    MINUS = 16
    PLUS = 17
    JCL_SR = 18
    JCL_SL = 19
    JCR_SR = 20
    JCR_SL = 21


class NXWrapper:
    def __init__(self, press_duration_ms: int = 50, delay_ms: int = 120):
        self.press_duration_ms = press_duration_ms
        self.delay_ms = delay_ms

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def button_hold(self, button_name: Button, duration_ms: int):
        """
        Hold a button for a certain duration.

        :param button_name: The name of the button to push.
        :param duration_ms: The duration in milliseconds.
        """
        pass

    def button_press(self, button_name: Button):
        """
        Press a button.

        :param button_name: The name of the button to push.
        """
        return self.button_hold(button_name, self.press_duration_ms)

    def series_press(self, button_names: list[Union[Button, tuple[Button, int]]]):
        """
        Press a series of buttons.
        :param button_names: A list of button names.
        """
        for button_name in button_names:
            if isinstance(button_name, tuple):
                self.button_hold(button_name[0], button_name[1])
            else:
                self.button_press(button_name)

    @abstractmethod
    def disconnect(self):
        pass

    # Context Manager
    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.disconnect()
        return False



import asyncio
import logging
import struct
import time
import socket
import math
from typing import Any
from contextlib import suppress

from joycontrol import utils
from joycontrol.my_semaphore import MyBoundedSemaphore

logger = logging.getLogger(__name__)


class NotConnectedError(ConnectionResetError):
    pass

class L2CAP_Transport(asyncio.Transport):
    def __init__(self, loop, protocol, itr_sock, ctr_sock, read_buffer_size, capture_file=None, flow_control = 4) -> None:
        super(L2CAP_Transport, self).__init__()

        self._loop = loop
        self._protocol = protocol

        self._itr_sock = itr_sock
        self._ctr_sock = ctr_sock

        self._capture_file = capture_file

        self._extra_info = {
            'peername': self._itr_sock.getpeername(),
            'sockname': self._itr_sock.getsockname(),
            'socket': self._itr_sock
        }

        self._is_closing = False

        # writing control
        self._write_lock = asyncio.Event()
        self._write_lock.set()
        self._write_lock_thread = utils.start_asyncio_thread(self._write_lock_monitor(), ignore=asyncio.CancelledError)
        self._write_window = MyBoundedSemaphore(flow_control)
        self._write_window_thread = utils.start_asyncio_thread(self._write_window_monitor(), ignore=asyncio.CancelledError)

        # reading control
        self._read_buffer_size = read_buffer_size
        self._is_reading = asyncio.Event()
        self._is_reading.set()
        self._read_thread = utils.start_asyncio_thread(self._reader(), ignore=asyncio.CancelledError)

    async def _write_window_monitor(self):
        with socket.socket(socket.AF_BLUETOOTH, socket.SOCK_RAW, socket.BTPROTO_HCI) as hci:
            hci.bind((0,))
            hci.setblocking(False)
            # 0x04 = HCI_EVT; 0x13 = Number of completed packets
            hci.setsockopt(socket.SOL_HCI, socket.HCI_FILTER, struct.pack("IIIh2x", 1 << 0x04, (1 << 0x13), 0, 0))

            while True:
                data = await self._loop.sock_recv(hci, 10)
                self._write_window.release(data[6] + data[7] * 0x100, best_effort=True)

    async def _write_lock_monitor(self):
        with socket.socket(socket.AF_BLUETOOTH, socket.SOCK_RAW, socket.BTPROTO_HCI) as hci:
            hci.bind((0,))
            hci.setblocking(False)
            # 0x04 = HCI_EVT; 0x1b = Max Slots Change
            hci.setsockopt(socket.SOL_HCI, socket.HCI_FILTER, struct.pack("IIIh2x", 1 << 0x04, (1 << 0x1b), 0, 0))
            while True:
                data = await self._loop.sock_recv(hci, 10)
                if data[5] < 5:
                    self.pause_writing()
                    await asyncio.sleep(1)
                    self.resume_writing()

    async def _reader(self):
        while True:
            await self._protocol.report_received(await self.read(), self._itr_sock.getpeername())

    async def read(self):
        """
        Read data from the underlying socket. This function waits,
        if reading is paused using the pause_reading function.

        :returns bytes
        """
        await self._is_reading.wait()
        data = await self._loop.sock_recv(self._itr_sock, self._read_buffer_size)

        # logger.debug(f'received "{list(data)}"')

        if not data:
            # disconnect happened
            logger.error('No data received.')
            self._protocol.connection_lost()

        if self._capture_file is not None:
            # write data to log file
            _time = struct.pack('d', time.time())
            size = struct.pack('i', len(data))
            self._capture_file.write(_time + size + data)

        return data

# Base transport API

    def write_eof():
        raise NotImplementedError("cannot write EOF")

    def get_extra_info(self, name: Any, default=None) -> Any:
        return self._extra_info.get(name, default)

    def is_closing(self) -> bool:
        return self._is_closing

    def set_protocol(self, protocol: asyncio.BaseProtocol) -> None:
        self._protocol = protocol

    def get_protocol(self) -> asyncio.BaseProtocol:
        return self._protocol

# Read-Transport API

    def is_reading(self) -> bool:
        """
        :returns True if the reader is running
        """
        return self._reader is not None and self._is_reading.is_set()

    def pause_reading(self) -> None:
        """
        Pauses any 'read' function calls.
        """
        self._is_reading.clear()

    def resume_reading(self) -> None:
        """
        Resumes all 'read' function calls.
        """
        self._is_reading.set()

    def set_read_buffer_size(self, size):
        self._read_buffer_size = size

# Write-Transport API:
# This is not compliant to the official trasnport API, as the core methods
# are asnyc. This is because the official API has no control over time and
# imho is quite lacking...

    def abort():
        raise NotImplementedError()

    def can_write_eof():
        return False

    def get_write_buffer_size():
        return self._write_window.get_aquired()

    def get_write_buffer_limits():
        return (0, self._write_window.get_limit())

    def set_write_buffer_limits(high=None, low=None):
        if low:
            raise NotImplementedError("Cannot set a lower bound for in flight data...")

        self._write_window.set_limit(high)

    async def write(self, data: Any) -> None:
        if isinstance(data, bytes):
            _bytes = data
        else:
            _bytes = bytes(data)

        if self._capture_file is not None:
            # write data to log file
            _time = struct.pack('d', time.time())
            size = struct.pack('i', len(_bytes))
            self._capture_file.write(_time + size + _bytes)

        # logger.debug(f'sending "{_bytes}"')

        try:
            await self._write_window.acquire()
            await self._write_lock.wait()
            await self._loop.sock_sendall(self._itr_sock, _bytes)
        except OSError as err:
            logger.error(err)
            self._protocol.connection_lost()
        except ConnectionResetError as err:
            logger.error(err)
            self._protocol.connection_lost()

    async def writelines(*data):
        for d in data:
            await self.write(data)

    def pause_writing(self):
        logger.info("pause transport write")
        self._write_lock.clear()

    def resume_writing(self):
        logger.info("resume transport write")
        self._write_lock.set()

    def is_writing(self):
        return not self._write_lock.is_set()

    async def close(self):
        """
        Stops reader and closes underlying socket
        """
        if not self._is_closing:
            # was not already closed
            self._is_closing = True

            self.pause_reading()
            self.pause_writing()

            self._read_thread.cancel()
            self._write_lock_thread.cancel()
            self._write_window_thread.cancel()

            with suppress(asyncio.CancelledError):
                await self._read_thread
            with suppress(asyncio.CancelledError):
                await self._write_lock_thread
            with suppress(asyncio.CancelledError):
                await self._write_window_thread

            # interrupt connection should be closed first
            self._itr_sock.close()
            self._ctr_sock.close()

            self._protocol.connection_lost(None)



import asyncio
import os

from .wrapper import Button, NXWrapper


class Mart1noJoyControl(NXWrapper):
    def __init__(self, press_duration_ms: int = 50):
        from .mart1no_joycontrol.joycontrol.controller import Controller
        from .mart1no_joycontrol.joycontrol.device import HidDevice
        from .mart1no_joycontrol.joycontrol.memory import FlashMemory
        from .mart1no_joycontrol.joycontrol.protocol import controller_protocol_factory
        from .mart1no_joycontrol.joycontrol.server import create_hid_server
        super().__init__(press_duration_ms)
        if not os.geteuid() == 0:
            raise PermissionError('Script must be run as root!')
        transport, protocol = asyncio.run(create_hid_server(
                controller_protocol_factory(Controller.from_arg("PRO_CONTROLLER"), spi_flash=FlashMemory()),
                ctl_psm=17,
                itr_psm=19
                )
                )
        controller_state = protocol.get_controller_state()
        self.transport = transport
        self.controller_state = controller_state

    def connect(self):
        asyncio.run(self.controller_state.connect())

    def button_hold(self, button: Button, duration_ms: int):
        from mart1no_joycontrol.joycontrol.controller_state import button_push
        # TODO get button name from enum
        raise NotImplementedError
        # asyncio.run(button_push(self.controller_state, button, sec=duration_ms / 1000))

    def disconnect(self):
        asyncio.run(self.transport.close())


class PoohlJoyControl(NXWrapper):
    def __init__(self, press_duration_ms: int = 50):
        from .Poohl_joycontrol.joycontrol.controller import Controller
        from .Poohl_joycontrol.joycontrol.device import HidDevice
        from .Poohl_joycontrol.joycontrol.memory import FlashMemory
        from .Poohl_joycontrol.joycontrol.protocol import controller_protocol_factory
        from .Poohl_joycontrol.joycontrol.server import create_hid_server
        super().__init__(press_duration_ms)
        if not os.geteuid() == 0:
            raise PermissionError('Script must be run as root!')
        hid = HidDevice(device_id=None)
        transport, protocol = asyncio.run(create_hid_server(
                controller_protocol_factory(Controller.from_arg("PRO_CONTROLLER"), spi_flash=FlashMemory()),
                ctl_psm=17,
                itr_psm=19
                )
                )
        controller_state = protocol.get_controller_state()
        self.transport = transport
        self.controller_state = controller_state

    def connect(self):
        asyncio.run(self.controller_state.connect())

    def button_hold(self, button: Button, duration_ms: int):
        from Poohl_joycontrol.joycontrol.controller_state import button_push
        # TODO get button name from enum
        raise NotImplementedError
        # asyncio.run(button_push(self.controller_state, button, sec=duration_ms / 1000))

    def disconnect(self):
        asyncio.run(self.transport.close())



import enum
from os import path as ph

import logging


logger = logging.getLogger(__name__)

unnamed_saves = 0

def get_savepath(hint='/tmp/amiibo'):
    global unnamed_saves
    unnamed_saves += 1
    if hint.endswith('.bin'):
        hint = hint[:-4]
    while True:
        path = hint + '_' + str(unnamed_saves) + '.bin'
        if not ph.exists(path):
            break
        unnamed_saves += 1
    return path


unnamed_backups = 0

def get_backuppath(hint='/tmp/amiibo.bin'):
    global unnamed_backups
    unnamed_backups += 1
    while True:
        path = hint + '.bak' + str(unnamed_backups)
        if not ph.exists(path):
            break
        unnamed_backups += 1
    return path


class NFCTagType(enum.Enum):
    AMIIBO = enum.auto


class NFCTag:
    """
    Class that represents a (Amiibo) NFC-Tag usually backed by a file. If needed files are generated.
    """
    def __init__(self, data, tag_type: NFCTagType = NFCTagType.AMIIBO, mutable=False, source=None):
        self.data: bytearray = bytearray(data)
        self.tag_type: NFCTagType = tag_type
        self.mutable: bool = mutable
        self.source: str = source
        if self.tag_type == NFCTagType.AMIIBO:
            if len(self.data) == 540:
                pass
            elif len(self.data) == 572:
                logger.info("Long amiibo loaded, manufacturer signature is ignored")
            else:
                logger.warning("Illegal Amiibo tag size")

    @classmethod
    def load_amiibo(cls, source):
        # if someone want to make this async have fun
        with open(source, "rb") as reader:
            return NFCTag(data=bytearray(reader.read()), tag_type=NFCTagType.AMIIBO, source=source)

    def create_backup(self):
        """
        copy the file backing this Tag
        """
        path = get_backuppath(self.source)
        logger.info(f"creating amiibo backup at {path}")
        with open(path, "wb") as writer:
            writer.write(self.data)

    def set_mutable(self, mutable=True):
        """
        By default tags are marked immutable to prevent corruption. To make them mutable create a backup first.
        @param mutable:
        @return:
        """
        if mutable > self.mutable:
            self.create_backup()
        self.mutable = mutable

    def save(self):
        if not self.source:
            self.source = get_savepath()
        with open(self.source, "wb") as writer:
            writer.write(self.data)
            logger.info(f"Saved altered amiibo as {self.source}")

    def getUID(self):
        return self.data[0:3] + self.data[4:8]

    def is_mutable(self):
        return self.mutable

    def write(self, idx, data):
        if idx > len(self.data) or idx+len(data) > len(self.data):
            logger.error(f"I Fucking hate pyhton {idx}, {bytes(data).hex()} {len(data)}")
        if not self.mutable:
            logger.warning("Ignored amiibo write to non-mutable amiibo")
        else:

            self.data[idx:idx + len(data)] = data




import asyncio
import logging
import time
from asyncio import BaseTransport, BaseProtocol
from typing import Optional, Union, Tuple, Text
import math

import enum
import joycontrol.debug as debug
import socket
import joycontrol.utils as utils
import struct
from joycontrol.controller import Controller
from joycontrol.controller_state import ControllerState
from joycontrol.memory import FlashMemory
from joycontrol.report import OutputReport, SubCommand, InputReport, OutputReportID
from joycontrol.transport import NotConnectedError
from joycontrol.mcu import MicroControllerUnit

logger = logging.getLogger(__name__)


def controller_protocol_factory(controller: Controller, spi_flash=None, reconnect = False):
    if isinstance(spi_flash, bytes):
        spi_flash = FlashMemory(spi_flash_memory_data=spi_flash)

    def create_controller_protocol():
        return ControllerProtocol(controller, spi_flash=spi_flash, reconnect = reconnect)

    return create_controller_protocol

class SwitchState(enum.Enum):
    STANDARD = enum.auto,
    GRIP_MENU = enum.auto,
    AWAITING_MAX_SLOTS = enum.auto

close_pairing_menu_map = {
    Controller.JOYCON_R: ['x', 'a', 'home'],
    Controller.JOYCON_L: ['down', 'left'],
    Controller.PRO_CONTROLLER: ['a', 'b', 'home']
}

close_pairing_masks = {
    Controller.JOYCON_R: int.from_bytes(bytes([0x2 | 0x8, 0x10, 0]), "big"),
    Controller.JOYCON_L: int.from_bytes(bytes([0, 0, 0x1 | 0x8]), "big"),
    Controller.PRO_CONTROLLER: int.from_bytes(bytes([0x4 | 0x8, 0x10, 0]), "big")
}

class ControllerProtocol(BaseProtocol):
    def __init__(self, controller: Controller, spi_flash: FlashMemory = None, reconnect = False):
        self.controller = controller
        self.spi_flash = spi_flash
        self.transport = None

        # time when the timer started.
        self._input_report_timer_start = None

        self._controller_state = ControllerState(self, controller, spi_flash=spi_flash)
        self._controller_state_sender = None
        self._writer_thread = None

        self._mcu = MicroControllerUnit(self._controller_state)

        self._is_pairing = not reconnect

        # input mode
        self.delay_map = {
            None: math.inf, # subcommands only
            0x3F: 1.0,
            0x21: math.inf, # shouldn't happen
            0x30: 1/60, # this needs revising, but 120 seems too fast
        #    0x30: 1/120 if self.controller == Controller.PRO_CONTROLLER else 1/60,
            0x31: 1/60
        }
        self._input_report_wakeup = asyncio.Event()
        self._set_mode(None)

        # "Pausing"-mechanism.
        self._not_paused = asyncio.Event()
        self._not_paused.set()

        self.sig_input_ready = asyncio.Event()
        self.sig_data_received = asyncio.Event()

# INTERNAL

    def _set_mode(self, mode, delay=None):

        if mode == 0x21:
            logger.error("Shouldn't go into subcommand mode")

        self._input_report_mode = mode
        if delay:
            self.send_delay = self.delay_map[mode]
        elif self._is_pairing:
            self.send_delay = 1/15
        elif mode in self.delay_map:
            self.send_delay = self.delay_map[mode]
        else:
            logger.warning(f"Unknown delay for mode {mode}, assuming 1/15")
            self.send_delay = 1/15

        if mode in [0x30, 0x31, 0x32, 0x33]:
            # sig input ready, writer
            pass

        self._input_report_wakeup.set()

    async def _write(self, input_report):
        """
        Fires sig_is_send event in the controller state afterwards.

        Raises NotConnected exception if the transport is not connected or the connection was lost.
        """
        if self.transport is None:
            raise NotConnectedError('Transport not registered.')

        if self._is_pairing and (int.from_bytes(input_report.data[4:7], "big") & close_pairing_masks[self.controller]):
            # this is a bit too early, but so far no
            logger.info('left change Grip/Order menu')
            self._is_pairing = False
            self._set_mode(self._input_report_mode)

        if not self._not_paused.is_set():
            logger.warning("Write while paused")

        await self.transport.write(input_report)

        self._controller_state.sig_is_send.set()

    def _generate_input_report(self, mode=None):
        input_report = InputReport()
        if not mode:
            mode = self._input_report_mode

        if not mode:
            raise ValueError("cannot generate Report without Mode")

        input_report.set_input_report_id(mode)
        if mode == 0x3F:
            input_report.data[1:3] = [0x28, 0xca, 0x08]
            input_report.data[4:11] = [0x40,0x8A, 0x4F, 0x8A, 0xD0, 0x7E, 0xDF, 0x7F]
        else:
            if self._input_report_timer_start:
                input_report.set_timer(round((time.time() - self._input_report_timer_start) / 0.005) % 0x100)
            else:
                input_report.set_timer(0)
            input_report.set_misc()
            input_report.set_button_status(self._controller_state.button_state)
            input_report.set_stick_status(self._controller_state.l_stick_state, self._controller_state.r_stick_state)
            input_report.set_vibrator_input()
            if mode == 0x21:
                pass # subcommand is set outside
            elif mode in [0x30, 0x31, 0x32, 0x33]:
                input_report.set_6axis_data()

            if mode == 0x31:
                input_report.set_ir_nfc_data(self._mcu.get_data())
        return input_report

    async def _writer(self):
        """
        This continuously sends input reports to the switch.
        This relies on the asyncio scheduler to sneak the additional
        subcommand-replies in
        """
        logger.info("writer started")
        while self.transport:
            await self._not_paused.wait()
            last_send_time = time.time()
            input_report = self._generate_input_report()
            try:
                await self._write(input_report)
            except:
                break
            # calculate delay
            self.send_delay = debug.get_delay(self.send_delay) #debug hook
            active_time = time.time() - last_send_time
            sleep_time = self.send_delay - active_time
            if sleep_time < 0:
                logger.warning(f'Code is running {abs(sleep_time)} s too slow!')
                sleep_time = 0

            try:
                await asyncio.wait_for(self._input_report_wakeup.wait(), timeout=sleep_time)
                self._input_report_wakeup.clear()
            except asyncio.TimeoutError as err:
                pass

        logger.warning("Writer exited...")
        return None

    async def _reply_to_sub_command(self, report):
        # classify sub command
        try:
            sub_command = report.get_sub_command()
        except NotImplementedError as err:
            logger.warning(err)
            return False

        if sub_command is None:
            raise ValueError('Received output report does not contain a sub command')

        logging.info(f'received Sub command {sub_command}')

        sub_command_data = report.get_sub_command_data()
        assert sub_command_data is not None

        response_report = self._generate_input_report(mode=0x21)

        try:
            # answer to sub command
            if sub_command == SubCommand.REQUEST_DEVICE_INFO:
                await self._command_request_device_info(response_report, sub_command_data)

            elif sub_command == SubCommand.SET_SHIPMENT_STATE:
                await self._command_set_shipment_state(response_report, sub_command_data)

            elif sub_command == SubCommand.SPI_FLASH_READ:
                await self._command_spi_flash_read(response_report, sub_command_data)

            elif sub_command == SubCommand.SET_INPUT_REPORT_MODE:
                await self._command_set_input_report_mode(response_report, sub_command_data)

            elif sub_command == SubCommand.TRIGGER_BUTTONS_ELAPSED_TIME:
                await self._command_trigger_buttons_elapsed_time(response_report, sub_command_data)

            elif sub_command == SubCommand.ENABLE_6AXIS_SENSOR:
                await self._command_enable_6axis_sensor(response_report, sub_command_data)

            elif sub_command == SubCommand.ENABLE_VIBRATION:
                await self._command_enable_vibration(response_report, sub_command_data)

            elif sub_command == SubCommand.SET_NFC_IR_MCU_CONFIG:
                await self._command_set_nfc_ir_mcu_config(response_report, sub_command_data)

            elif sub_command == SubCommand.SET_NFC_IR_MCU_STATE:
                await self._command_set_nfc_ir_mcu_state(response_report, sub_command_data)

            elif sub_command == SubCommand.SET_PLAYER_LIGHTS:
                await self._command_set_player_lights(response_report, sub_command_data)
            else:
                logger.warning(f'Sub command 0x{sub_command.value:02x} not implemented - ignoring')
                return False

            await self._write(response_report)

        except NotImplementedError as err:
            logger.error(f'Failed to answer {sub_command} - {err}')
            return False
        return True

# transport hooks

    def connection_made(self, transport: BaseTransport) -> None:
        logger.debug('Connection established.')
        self.transport = transport
        self._input_report_timer_start = time.time()

    def connection_lost(self, exc: Optional[Exception] = None) -> None:
        if self.transport is not None:
            logger.error('Connection lost.')
            asyncio.ensure_future(self.transport.close())
            self.transport = None

            if self._controller_state_sender is not None:
                self._controller_state_sender.set_exception(NotConnectedError)

    def error_received(self, exc: Exception) -> None:
        # TODO?
        raise NotImplementedError()

    async def report_received(self, data: Union[bytes, Text], addr: Tuple[str, int]) -> None:
        self.sig_data_received.set()

        try:
            report = OutputReport(list(data))
        except ValueError as v_err:
            logger.warning(f'Report parsing error "{v_err}" - IGNORE')
            return

        try:
            output_report_id = report.get_output_report_id()
        except NotImplementedError as err:
            logger.warning(err)
            return

        if output_report_id == OutputReportID.SUB_COMMAND:
            await self._reply_to_sub_command(report)
        elif output_report_id == OutputReportID.RUMBLE_ONLY:
            # TODO Rumble
            pass
        elif output_report_id == OutputReportID.REQUEST_IR_NFC_MCU:
            self._mcu.received_11(report.data[11], report.get_sub_command_data())
            pass
        else:
            logger.warning(f'Output report {output_report_id} not implemented - ignoring')


# event lisnter hooks

    async def send_controller_state(self):
        """
        Waits for the controller state to be send.

        Raises NotConnected exception if the transport is not connected or the connection was lost.
        """
        # TODO: Call write directly if in continuously sending input report mode

        if self.transport is None:
            raise NotConnectedError('Transport not registered.')

        if not self._not_paused.is_set():
            await self._write(self._generate_input_report())
        else:
            self._controller_state.sig_is_send.clear()

            # wrap into a future to be able to set an exception in case of a disconnect
            self._controller_state_sender = asyncio.ensure_future(self._controller_state.sig_is_send.wait())
            await self._controller_state_sender
            self._controller_state_sender = None

    async def wait_for_output_report(self):
        """
        Waits until an output report from the Switch is received.
        """
        self.sig_data_received.clear()
        await self.sig_data_received.wait()

    def pause(self):
        logger.info("paused")
        self._not_paused.clear()

    def unpause(self):
        logger.info("unpaused")
        self._not_paused.set()

    def get_controller_state(self) -> ControllerState:
        return self._controller_state

# subcommands

    async def _command_request_device_info(self, input_report, sub_command_data):

        address = self.transport.get_extra_info('sockname')
        assert address is not None
        bd_address = list(map(lambda x: int(x, 16), address[0].split(':')))

        input_report.set_ack(0x82)
        input_report.sub_0x02_device_info(bd_address, controller=self.controller)

        return input_report

    async def _command_set_shipment_state(self, input_report, sub_command_data):
        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(0x08)
        return input_report

    async def _command_spi_flash_read(self, input_report, sub_command_data):
        """
        Replies with 0x21 input report containing requested data from the flash memory.
        :param sub_command_data: input report sub command data bytes
        """
        input_report.set_ack(0x90)

        # parse offset
        offset = 0
        digit = 1
        for i in range(4):
            offset += sub_command_data[i] * digit
            digit *= 0x100

        size = sub_command_data[4]

        if self.spi_flash is not None:
            spi_flash_data = self.spi_flash[offset: offset + size]
            input_report.sub_0x10_spi_flash_read(offset, size, spi_flash_data)
        else:
            spi_flash_data = size * [0x00]
            input_report.sub_0x10_spi_flash_read(offset, size, spi_flash_data)

        return input_report

    async def _command_set_input_report_mode(self, input_report, sub_command_data):
        if self._input_report_mode == sub_command_data[0]:
            logger.warning(f'Already in input report mode {sub_command_data[0]} - ignoring request')

        self._set_mode(sub_command_data[0])

        # Send acknowledgement

        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(0x03)

        return input_report

    async def _command_trigger_buttons_elapsed_time(self, input_report, sub_command_data):
        input_report.set_ack(0x83)
        input_report.reply_to_subcommand_id(SubCommand.TRIGGER_BUTTONS_ELAPSED_TIME)
        # Hack: We assume this command is only used during pairing - Set values so the Switch assigns a player number
        if self.controller == Controller.PRO_CONTROLLER:
            input_report.sub_0x04_trigger_buttons_elapsed_time(L_ms=3000, R_ms=3000)
        elif self.controller in (Controller.JOYCON_L, Controller.JOYCON_R):
            # TODO: What do we do if we want to pair a combined JoyCon?
            input_report.sub_0x04_trigger_buttons_elapsed_time(SL_ms=3000, SR_ms=3000)
        else:
            raise NotImplementedError(self.controller)

        return input_report

    async def _command_enable_6axis_sensor(self, input_report, sub_command_data):
        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(0x40)

        return input_report

    async def _command_enable_vibration(self, input_report, sub_command_data):
        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(SubCommand.ENABLE_VIBRATION.value)

        return input_report

    async def _command_set_nfc_ir_mcu_config(self, input_report, sub_command_data):
        self._mcu.set_config_cmd(sub_command_data)

        input_report.set_ack(0xA0)
        input_report.reply_to_subcommand_id(SubCommand.SET_NFC_IR_MCU_CONFIG.value)

        data = [1, 0, 255, 0, 8, 0, 27, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 200]
        for i in range(len(data)):
            input_report.data[16 + i] = data[i]

        return input_report

    async def _command_set_nfc_ir_mcu_state(self, input_report, sub_command_data):
        self._mcu.set_power_state_cmd(sub_command_data[0])

        if sub_command_data[0] == 0x01:
            # 0x01 = Resume
            input_report.set_ack(0x80)
            input_report.reply_to_subcommand_id(SubCommand.SET_NFC_IR_MCU_STATE.value)
        elif sub_command_data[0] == 0x00:
            # 0x00 = Suspend
            input_report.set_ack(0x80)
            input_report.reply_to_subcommand_id(SubCommand.SET_NFC_IR_MCU_STATE.value)
        else:
            raise NotImplementedError(f'Argument {sub_command_data[0]} of {SubCommand.SET_NFC_IR_MCU_STATE} '
                                      f'not implemented.')
        return input_report

    async def _command_set_player_lights(self, input_report, sub_command_data):
        input_report.set_ack(0x80)
        input_report.reply_to_subcommand_id(SubCommand.SET_PLAYER_LIGHTS.value)

        self._writer_thread = utils.start_asyncio_thread(self._writer())
        self.sig_input_ready.set()
        return input_report



import asyncio
import logging
import struct
import time
from typing import Any

from joycontrol import utils

logger = logging.getLogger(__name__)


class NotConnectedError(ConnectionResetError):
    pass


class L2CAP_Transport(asyncio.Transport):
    def __init__(self, loop, protocol, itr_sock, ctr_sock, read_buffer_size, capture_file=None) -> None:
        super(L2CAP_Transport, self).__init__()

        self._loop = loop
        self._protocol = protocol

        self._itr_sock = itr_sock
        self._ctr_sock = ctr_sock

        self._read_buffer_size = read_buffer_size

        self._extra_info = {
            'peername': self._itr_sock.getpeername(),
            'sockname': self._itr_sock.getsockname(),
            'socket': self._itr_sock
        }

        self._is_closing = False
        self._is_reading = asyncio.Event()

        self._capture_file = capture_file

        # start underlying reader
        self._read_thread = None
        self._is_reading.set()
        self.start_reader()

    async def _reader(self):
        while True:
            try:
                data = await self.read()
            except NotConnectedError:
                self._read_thread = None
                break

            await self._protocol.report_received(data, self._itr_sock.getpeername())

    def start_reader(self):
        """
        Starts the transport reader which calls the protocols report_received function for every incoming message
        """
        if self._read_thread is not None:
            raise ValueError('Reader is already running.')

        self._read_thread = asyncio.ensure_future(self._reader())

        # Create callback in case the reader is failing
        callback = utils.create_error_check_callback(ignore=asyncio.CancelledError)
        self._read_thread.add_done_callback(callback)

    async def set_reader(self, reader: asyncio.Future):
        """
        Cancel the currently running reader and register the new one.
        A reader is a coroutine that calls this transports 'read' function.
        The 'read' function calls can be paused by calling pause_reading of this transport.
        :param reader: future reader
        """
        if self._read_thread is not None:
            # cancel currently running reader
            if self._read_thread.cancel():
                try:
                    await self._read_thread
                except asyncio.CancelledError:
                    pass

        # Create callback for debugging in case the reader is failing
        err_callback = utils.create_error_check_callback(ignore=asyncio.CancelledError)
        reader.add_done_callback(err_callback)

        self._read_thread = reader

    def get_reader(self):
        return self._read_thread

    async def read(self):
        """
        Read data from the underlying socket. This function waits,
        if reading is paused using the pause_reading function.

        :returns bytes
        """
        await self._is_reading.wait()
        data = await self._loop.sock_recv(self._itr_sock, self._read_buffer_size)

        # logger.debug(f'received "{list(data)}"')

        if not data:
            # disconnect happened
            logger.error('No data received.')
            self._protocol.connection_lost()
            raise NotConnectedError('No data received.')

        if self._capture_file is not None:
            # write data to log file
            _time = struct.pack('d', time.time())
            size = struct.pack('i', len(data))
            self._capture_file.write(_time + size + data)

        return data

    def is_reading(self) -> bool:
        """
        :returns True if the reader is running
        """
        return self._reader is not None and self._is_reading.is_set()

    def pause_reading(self) -> None:
        """
        Pauses any 'read' function calls.
        """
        self._is_reading.clear()

    def resume_reading(self) -> None:
        """
        Resumes all 'read' function calls.
        """
        self._is_reading.set()

    def set_read_buffer_size(self, size):
        self._read_buffer_size = size

    async def write(self, data: Any) -> None:
        if isinstance(data, bytes):
            _bytes = data
        else:
            _bytes = bytes(data)

        if self._capture_file is not None:
            # write data to log file
            _time = struct.pack('d', time.time())
            size = struct.pack('i', len(_bytes))
            self._capture_file.write(_time + size + _bytes)

        # logger.debug(f'sending "{_bytes}"')

        try:
            await self._loop.sock_sendall(self._itr_sock, _bytes)
        except OSError as err:
            logger.error(err)
            self._protocol.connection_lost()
            raise NotConnectedError(err)
        except ConnectionResetError as err:
            logger.error(err)
            self._protocol.connection_lost()
            raise err

    def abort(self) -> None:
        raise NotImplementedError

    def get_extra_info(self, name: Any, default=None) -> Any:
        return self._extra_info.get(name, default)

    def is_closing(self) -> bool:
        return self._is_closing

    async def close(self):
        """
        Stops reader and closes underlying socket
        """
        if not self._is_closing:
            # was not already closed
            self._is_closing = True
            if self._read_thread.cancel():
                # wait for reader to cancel
                try:
                    await self._read_thread
                except asyncio.CancelledError:
                    pass

            # interrupt connection should be closed first
            self._itr_sock.close()
            self._ctr_sock.close()

    def set_protocol(self, protocol: asyncio.BaseProtocol) -> None:
        self._protocol = protocol

    def get_protocol(self) -> asyncio.BaseProtocol:
        return self._protocol
